"""Contains the MpfMc base class, which is the main App instance for the
mpf-mc."""
import importlib
import os
import queue
import sys
import threading
import time
import logging

import gc
import weakref

from typing import Dict

from kivy.app import App
from kivy.clock import Clock
from kivy.resources import resource_add_path
from kivy.logger import Logger

from kivy.config import Config

import mpf
from mpf._version import __version__ as __mpfversion__
from mpf.core.config_loader import MpfMcConfig
from mpf.core.case_insensitive_dict import CaseInsensitiveDict
from mpf.core.config_validator import ConfigValidator
from mpf.core.events import EventManager
from mpf.core.player import Player
from mpf.core.device_manager import DeviceCollection
from mpf.core.utility_functions import Util
from mpf.core.rgb_color import RGBColor

import mpfmc
from mpfmc._version import __version__
from mpfmc.assets.video import VideoAsset
from mpfmc.core.bcp_processor import BcpProcessor
from mpfmc.core.config_processor import ConfigProcessor
from mpfmc.core.mode_controller import ModeController
from mpfmc.uix.transitions import TransitionManager
from mpfmc.uix.effects import EffectsManager
from mpfmc.core.config_collection import create_config_collections
from mpfmc.assets.image import ImageAsset
from mpfmc.assets.bitmap_font import BitmapFontAsset
from mpfmc.core.dmd import Dmd, RgbDmd
from mpfmc.core.assets import ThreadedAssetManager
from mpfmc.core.mc_placeholder_manager import McPlaceholderManager
from mpfmc.core.mc_settings_controller import McSettingsController

try:
    from mpfmc.core.audio import SoundSystem
    from mpfmc.assets.sound import SoundAsset
except ImportError:
    SoundSystem = None
    SoundAsset = None
    logging.warning("mpfmc.core.audio library could not be loaded. Audio "
                    "features will not be available")

# The following line is needed to allow mpfmc modules to use the
# getLogger(name) method
logging.Logger.manager.root = Logger


# pylint: disable-msg=too-many-instance-attributes
class MpfMc(App):

    """Kivy app for the mpf media controller."""

    # pylint: disable-msg=too-many-statements
    def __init__(self, options, config: MpfMcConfig,
                 thread_stopper=None):

        self.log = logging.getLogger('mpfmc')
        self.log.info("Mission Pinball Framework Media Controller v%s", __version__)
        self.log.info("Mission Pinball Framework Game Engine v%s", __mpfversion__)

        if (__version__.split('.')[0] != __mpfversion__.split('.')[0] or
                __version__.split('.')[1] != __mpfversion__.split('.')[1]):

            self.log.error("MPF MC and MPF Game engines must be same "
                           "major.minor versions. You have MPF v%s and MPF-MC"
                           " v%s", __mpfversion__, __version__)

            raise ValueError("MPF MC and MPF Game engines must be same "
                             "major.minor versions. You have MPF v{} and MPF-MC"
                             " v{}".format(__mpfversion__, __version__))

        super().__init__()

        self.options = options
        self.machine_path = config.get_machine_path()
        self.log.info("Machine path: %s", self.machine_path)

        # load machine into path to load modules
        if self.machine_path not in sys.path:
            sys.path.append(self.machine_path)
        self.mc_config = config
        self.config_validator = ConfigValidator(self, config.get_config_spec())
        self.machine_config = self.mc_config.get_machine_config()
        self.config = self.machine_config

        self.clock = Clock
        # pylint: disable-msg=protected-access
        self.log.info("Starting clock at %sHz", Clock._max_fps)
        self._boot_holds = set()
        self.is_init_done = threading.Event()
        self.mpf_path = os.path.dirname(mpf.__file__)
        self.modes = CaseInsensitiveDict()
        self.player_list = list()
        self.player = None
        self.num_players = 0
        self.bcp_client_connected = False
        self.placeholder_manager = McPlaceholderManager(self)
        self.settings = McSettingsController(self)

        self.animation_configs = dict()
        self.active_slides = dict()
        self.custom_code = list()

        self.register_boot_hold('init')
        self.displays = DeviceCollection(self, "displays", "displays")
        self.machine_vars = CaseInsensitiveDict()
        self.machine_var_monitor = False
        self.monitors = dict()
        self.targets = dict()
        """Dict which contains all the active slide frames in the machine that
        a slide can target. Will always contain an entry called 'default'
        which will be used if a slide doesn't specify targeting.
        """

        self.keyboard = None
        self.dmds = []
        self.rgb_dmds = []
        self.crash_queue = queue.Queue()
        self.ticks = 0
        self.start_time = 0
        self.debug_refs = []

        MYPY = False    # NOQA
        if MYPY:  # pragma: no cover
            self.videos = None               # type: Dict[str, VideoAsset]

        if thread_stopper:
            self.thread_stopper = thread_stopper
        else:
            self.thread_stopper = threading.Event()

        # Core components
        self.events = EventManager(self)
        self.mode_controller = ModeController(self)
        create_config_collections(self, self.machine_config['mpf-mc']['config_collections'])
        self._preprocess_config(self.config)

        self.config_processor = ConfigProcessor(self)
        self.transition_manager = TransitionManager(self)
        self.effects_manager = EffectsManager(self)

        self._set_machine_path()

        self._load_font_paths()

        # Initialize the sound system (must be done prior to creating the AssetManager).
        # If the sound system is not available, do not load any other sound-related modules.
        if SoundSystem is None or self.options.get("no_sound"):
            self.sound_system = None
        else:
            self.sound_system = SoundSystem(self)
            if self.sound_system.audio_interface is None:
                self.sound_system = None

        self.asset_manager = ThreadedAssetManager(self)
        self.bcp_processor = BcpProcessor(self)

        # Asset classes
        ImageAsset.initialize(self)
        VideoAsset.initialize(self)
        BitmapFontAsset.initialize(self)

        self._initialise_sound_system()

        self.clock.schedule_interval(self._check_crash_queue, 1)

        self.events.add_handler("client_connected", self._create_dmds)
        self.events.add_handler("player_turn_start", self.player_start_turn)

        self.create_machine_var('mpfmc_ver', __version__)
        # force setting it here so we have it before MPF connects
        self.receive_machine_var_update('mpfmc_ver', __version__, 0, True)

    def _load_named_colors(self):
        for name, color in self.machine_config.get('named_colors', {}).items():
            RGBColor.add_color(name, color)

    def track_leak_reference(self, element):
        """Track elements to find leaks."""
        if not self.options["production"]:
            self.debug_refs.append(weakref.ref(element))
            # cleanup all dead references
            self.debug_refs = [element for element in self.debug_refs if element()]

    @staticmethod
    def _preprocess_config(config):
        kivy_config = config['kivy_config']

        try:
            kivy_config['graphics'].update(config['window'])
        except KeyError:
            pass

        if ('top' in kivy_config['graphics'] and
                'left' in kivy_config['graphics']):
            kivy_config['graphics']['position'] = 'custom'

        for section, settings in kivy_config.items():
            for k, v in settings.items():
                try:
                    if k in Config[section]:
                        Config.set(section, k, v)
                except KeyError:
                    continue

        try:  # config not validated yet, so we use try
            if config['window']['exit_on_escape']:
                Config.set('kivy', 'exit_on_escape', '1')
        except KeyError:
            pass

        Config.set('graphics', 'maxfps', int(config['mpf-mc']['fps']))

    def _load_config(self):
        files = [os.path.join(
            mpfmc.__path__[0], self.options["mcconfigfile"])]
        for config_file in self.options["configfile"]:
            files.append(os.path.join(self.machine_path, "config", config_file))
        mpf_config = self.mpf_config_processor.load_config_files_with_cache(files, "machine", True)

        self._preprocess_config(mpf_config)

        return mpf_config

    def _create_dmds(self, **kwargs):
        del kwargs
        self.create_dmds()
        self.create_rgb_dmds()
        self.events.remove_all_handlers_for_event("client_connected")

    def _load_font_paths(self):
        # Add local machine fonts path
        if os.path.isdir(os.path.join(self.machine_path,
                                      self.machine_config['mpf-mc']['paths']['fonts'])):

            resource_add_path(os.path.join(self.machine_path,
                                           self.machine_config['mpf-mc']['paths']['fonts']))

        # Add mpfmc fonts path
        resource_add_path(os.path.join(os.path.dirname(mpfmc.__file__),
                                       'fonts'))

    def _initialise_sound_system(self):
        # Only initialize sound assets if sound system is loaded and enabled
        if self.sound_system is not None and self.sound_system.enabled:
            SoundAsset.extensions = tuple(
                self.sound_system.audio_interface.supported_extensions())
            SoundAsset.initialize(self)
        else:
            # If the sound system is not loaded or enabled, remove the
            # audio-related config_player modules and config collections
            del self.machine_config['mpf-mc']['config_players']['sound']
            del self.machine_config['mpf-mc']['config_players']['track']
            del self.machine_config['mpf-mc']['config_players']['sound_loop']
            del self.machine_config['mpf-mc']['config_players']['playlist']
            del self.machine_config['mpf-mc']['config_collections']['sound_loop_set']
            del self.machine_config['mpf-mc']['config_collections']['playlist']

    def get_system_config(self):
        return self.machine_config['mpf-mc']

    def validate_machine_config_section(self, section):
        """Validate machine config."""
        if section not in self.config_validator.get_config_spec():
            return

        if section not in self.machine_config:
            self.machine_config[section] = dict()

        self.machine_config[section] = self.config_validator.validate_config(
            section, self.machine_config[section], section)

    def get_config(self):
        return self.machine_config

    def _set_machine_path(self):
        self.log.debug("Machine path: %s", self.machine_path)

        # Add the machine folder to sys.path so we can import modules from it
        sys.path.insert(0, self.machine_path)

    def register_boot_hold(self, hold):
        # print('registering boot hold', hold)
        if self.is_init_done.is_set():
            raise AssertionError("Register hold after init_done")
        self._boot_holds.add(hold)

    def clear_boot_hold(self, hold):
        if self.is_init_done.is_set():
            raise AssertionError("Register hold after init_done")
        self._boot_holds.remove(hold)
        # print('clearing boot hold', hold, self._boot_holds)
        self.log.debug('Clearing boot hold %s. Holds remaining: %s', hold, self._boot_holds)
        if not self._boot_holds:
            self.init_done()

    def _register_config_players(self):
        # todo move this to config_player module

        for name, module in self.machine_config['mpf-mc'][
                'config_players'].items():
            imported_module = importlib.import_module(module)
            setattr(self, '{}_player'.format(name),
                    imported_module.mc_player_cls(self))

    def displays_initialized(self, *args):
        del args
        self.validate_machine_config_section('window')
        from mpfmc.uix.window import Window
        Window.initialize(self)
        self.events.post('displays_initialized')
        '''event: displays_initialized
        desc: Posted as soon as MPF MC displays have been initialized.

        Note that this event is used as part of the internal MPF-MC startup
        process. In some cases it will be posted *before* the slide_player is
        ready, meaning that you *CANNOT* use this event to post slides or play
        sounds.

        Instead, use the *mc_ready* event, which is posted as early as possible
        once the slide player and sound players are setup.

        Note that this event is generated by the media controller and does not
        exist on the MPF side of things.

        Also note that if you're using a media controller other than the MPF-MC
        (such as the Unity 3D backbox controller), then this event won't exist.

        '''
        self.events.process_event_queue()
        self.events.remove_all_handlers_for_event("displays_initialized")
        self._init()

    def create_dmds(self):
        """Create DMDs."""
        if 'dmds' in self.machine_config:
            for name, config in self.machine_config['dmds'].items():
                dmd = Dmd(self, name, config)
                self.dmds.append(dmd)

    def create_rgb_dmds(self):
        """Create RBG DMDs."""
        if 'rgb_dmds' in self.machine_config:
            for name, config in self.machine_config['rgb_dmds'].items():
                dmd = RgbDmd(self, name, config)
                self.rgb_dmds.append(dmd)

    def _init(self):
        # Since the window is so critical in Kivy, we can't continue the
        # boot process until the window is setup, and we can't set the
        # window up until the displays are initialized.

        self._load_named_colors()
        self._register_config_players()
        self.events.post("init_phase_1")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()
        self.events.post("mc_ready")
        '''event: mc_ready
        desc: Posted when the MPF-MC is available to start showing slides and
        playing sounds.

        Note that this event does not mean the MC is done loading. Instead it's
        posted at the earliest possible moment that the core MC components are
        available, meaning you can trigger "boot" slides from this event (which
        could in turn be used to show asset loading status, boot progress,
        etc.)

        If you want to show slides that require images or video loaded from
        disk, use the event "init_done" instead which is posted once all the
        assets set to "preload" have been loaded.
        '''

        self.events.process_event_queue()
        self.events.post("init_phase_2")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()
        self.events.post("init_phase_3")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()
        self._load_custom_code()
        self.events.post("init_phase_4")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()
        self.events.post("init_phase_5")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()
        self.clear_boot_hold('init')
        self.events.remove_all_handlers_for_event("init_phase_1")
        self.events.remove_all_handlers_for_event("init_phase_2")
        self.events.remove_all_handlers_for_event("init_phase_3")
        self.events.remove_all_handlers_for_event("init_phase_4")
        self.events.remove_all_handlers_for_event("init_phase_5")

    def init_done(self):
        self.is_init_done.set()
        self.events.post("init_done")
        # no events docstring as this event is also in mpf
        self.events.process_event_queue()

    def build(self):
        self.start_time = time.time()
        self.ticks = 0
        self.clock.schedule_interval(self.tick, 0)
        self.events.add_handler("debug_dump_stats", self._debug_dump_displays)

    def _debug_dump_displays(self, **kwargs):
        del kwargs
        self.log.info("--- DEBUG DUMP DISPLAYS ---")
        self.log.info("Active slides: %s (Count: %s). Displays: %s (Count: %s). Available Slides: %s",
                      self.active_slides, len(self.active_slides), self.displays, len(self.displays), len(self.slides))
        for display in self.displays:
            self.log.info("Listing children for display: %s", display)
            children = 0
            for child in display.walk():
                self.log.info(child)
                children += 1
            self.log.info("Total children: %s", children)
        self.log.info("--- DEBUG DUMP DISPLAYS END ---")
        gc.collect()
        if not self.options["production"]:
            self.log.info("--- DEBUG DUMP OBJECTS ---")
            self.log.info("Elements in list (may be dead): %s", len(self.debug_refs))
            for element in self.debug_refs:
                real_element = element()
                if real_element:
                    self.log.info(real_element)
            self.log.info("--- DEBUG DUMP OBJECTS END ---")
        else:
            self.log.info("--- DEBUG DUMP OBJECTS DISABLED BECAUSE OF PRODUCTION FLAG ---")
        self.log.info("--- DEBUG DUMP CLOCK ---")
        ev = Clock._root_event  # pylint: disable-msg=protected-access
        while ev:
            self.log.info(ev)
            ev = ev.next
        self.log.info("--- DEBUG DUMP CLOCK END ---")

    def on_stop(self):
        self.log.info("Stopping...")
        self.thread_stopper.set()

        self.events.post("shutdown")
        self.events.process_event_queue()

        try:
            self.log.info("Loop rate %s Hz", round(self.ticks / (time.time() - self.start_time), 2))
        except ZeroDivisionError:
            pass

    def reset(self, **kwargs):
        del kwargs
        self.player = None
        self.player_list = list()

        self.events.post('mc_reset_phase_1')
        '''event: mc_reset_phase_1
        desc: Posted on the MPF-MC only (e.g. not in MPF). This event is used
        internally as part of the MPF-MC reset process.
        '''
        self.events.process_event_queue()
        self.events.post('mc_reset_phase_2')
        '''event: mc_reset_phase_2
        desc: Posted on the MPF-MC only (e.g. not in MPF). This event is used
        internally as part of the MPF-MC reset process.
        '''
        self.events.process_event_queue()
        self.events.post('mc_reset_phase_3')
        '''event: mc_reset_phase_3
        desc: Posted on the MPF-MC only (e.g. not in MPF). This event is used
        internally as part of the MPF-MC reset process.
        '''
        self.events.process_event_queue()
        self.events.post('mc_reset_complete')
        '''event: mc_reset_complete
        desc: Posted on the MPF-MC only (e.g. not in MPF). This event is posted
        when the MPF-MC reset process is complete.
        '''

    def game_start(self, **kargs):
        self.player = None
        self.player_list = list()
        self.num_players = 0
        self.events.post('game_started', **kargs)
        # no events docstring as this event is also in mpf

    def game_end(self, **kwargs):
        self.player = None
        self.events.post('game_ended', **kwargs)
        # no events docstring as this event is also in mpf

    def add_player(self, player_num):
        if player_num > len(self.player_list):
            player = Player(self, len(self.player_list))
            self.player_list.append(player)

            self.events.post('player_added', player=player, num=player_num)
            # no events docstring as this event is also in mpf

            # Enable player var events and send all initial values
            player.enable_events(True, True)

    def update_player_var(self, name, value, player_num):
        try:
            self.player_list[int(player_num) - 1][name] = value
        except (IndexError, KeyError):
            pass

    def player_start_turn(self, number, **kwargs):
        del kwargs
        if ((self.player and self.player.number != number) or
                not self.player):

            try:
                self.player = self.player_list[int(number) - 1]
                self.events.post('player_turn_start', number=number,
                                 player=self.player)
            except IndexError:
                self.log.error('Received player turn start for player %s, but '
                               'only %s player(s) exist',
                               number, len(self.player_list))

    def create_machine_var(self, name, value):
        """Same as set_machine_var."""
        self.set_machine_var(name, value)

    def set_machine_var(self, name, value):
        """Set machine var and send it via BCP to MPF."""
        if hasattr(self, "bcp_processor") and self.bcp_processor.connected:
            self.bcp_processor.send_machine_var_to_mpf(name, value)

    def receive_machine_var_update(self, name, value, change, prev_value):
        """Update a machine var received via BCP."""
        if value is None:
            try:
                del self.machine_vars[name]
            except KeyError:
                pass
        else:
            self.machine_vars[name] = value

        if change:
            self.log.debug("Setting machine_var '%s' to: %s, (prior: %s, "
                           "change: %s)", name, value, prev_value,
                           change)
            self.events.post('machine_var_' + name,
                             value=value,
                             prev_value=prev_value,
                             change=change)
            # no events docstring as this event is also in mpf

    def tick(self, dt):
        """Process event queue."""
        del dt
        self.ticks += 1
        self.events.process_event_queue()

    def _load_custom_code(self):
        if 'mc_scriptlets' in self.machine_config:
            self.machine_config['mc_scriptlets'] = (
                self.machine_config['mc_scriptlets'].split(' '))

            self.log.debug("Loading scriptlets... (deprecated)")

            for scriptlet in self.machine_config['mc_scriptlets']:

                self.log.debug("Loading '%s' scriptlet (deprecated)", scriptlet)

                scriptlet_obj = Util.string_to_class(
                    self.machine_config['mpf-mc']['paths']['scriptlets'] +
                    "." + scriptlet)(mc=self,
                                     name=scriptlet.split('.')[1])

                self.custom_code.append(scriptlet_obj)

        if 'mc_custom_code' in self.machine_config:
            self.log.debug("Loading custom_code...")

            for custom_code in self.machine_config['mc_custom_code']:

                self.log.debug("Loading '%s' custom_code", custom_code)

                custom_code_obj = Util.string_to_class(
                    self.machine_config['mpf-mc']['paths']['scriptlets'] +
                    "." + custom_code)(mc=self,
                                       name=custom_code.split('.')[1])

                self.custom_code.append(custom_code_obj)

    def _check_crash_queue(self, dt):
        del dt
        try:
            crash = self.crash_queue.get(block=False)
        except queue.Empty:
            pass
        else:
            self.log.critical("Shutting down due to child thread crash")
            self.log.critical("Crash details: %s", crash)
            self.stop()

    def register_monitor(self, monitor_class, monitor):
        """Registers a monitor.

        Args:
            monitor_class: String name of the monitor class for this monitor
                that's being registered.
            monitor: String name of the monitor.

        MPF uses monitors to allow components to monitor certain internal
        elements of MPF.

        For example, a player variable monitor could be setup to be notified of
        any changes to a player variable, or a switch monitor could be used to
        allow a plugin to be notified of any changes to any switches.

        The MachineController's list of registered monitors doesn't actually
        do anything. Rather it's a dictionary of sets which the monitors
        themselves can reference when they need to do something. We just needed
        a central registry of monitors.

        """
        if monitor_class not in self.monitors:
            self.monitors[monitor_class] = set()

        self.monitors[monitor_class].add(monitor)

    def post_mc_native_event(self, event, **kwargs):
        if self.bcp_processor.enabled and self.bcp_client_connected:
            self.bcp_processor.send('trigger', name=event, **kwargs)

        self.events.post(event, **kwargs)



from functools import partial

from kivy.graphics.instructions import Callback
from kivy.uix.relativelayout import RelativeLayout

from kivy.clock import Clock
from kivy.graphics.fbo import Fbo
from kivy.graphics.opengl import glReadPixels, GL_RGBA, GL_UNSIGNED_BYTE
from kivy.graphics.texture import Texture

from mpfmc.core.bcp_config_player import BcpConfigPlayer


class McDisplayLightPlayer(BcpConfigPlayer):

    """Grabs pixel from a display and use them as lights."""

    config_file_section = 'display_light_player'
    show_section = 'display_lights'
    machine_collection_name = 'displays'

    def __init__(self, machine):
        super().__init__(machine)
        self._scheduled = False
        self._last_color = {}

    # pylint: disable-msg=too-many-arguments
    def play_element(self, settings, element, context, calling_context, priority=0, **kwargs):
        context_dict = self._get_instance_dict(context)
        if settings['action'] == "play":
            if not self._scheduled:
                self._scheduled = True
                Clock.schedule_interval(self._tick, 0)
            if element not in context_dict:
                context_dict[element] = self._setup_fbo(element, settings, context)
            else:
                context_dict[element][5] = True
        elif settings['action'] == "stop":
            try:
                context_dict[element][5] = False
            except IndexError:
                pass
        else:
            raise AssertionError("Unknown action {}".format(settings['action']))

    def _setup_fbo(self, element, settings, context):
        """Setup FBO for a display."""
        if element not in self.machine.displays:
            raise AssertionError("Display {} not found. Please create it to use display_light_player.".format(element))
        source = self.machine.displays[element]

        # put the widget canvas on a Fbo
        texture = Texture.create(size=source.size, colorfmt='rgba')
        fbo = Fbo(size=source.size, texture=texture)

        effect_widget = RelativeLayout()

        effect_widget.size = source.size

        fbo.add(effect_widget.canvas)
        with source.canvas:
            callback = Callback(partial(self._trigger_render, context, element))

        return [fbo, effect_widget, source, settings, True, True, True, callback]

    def _trigger_render(self, context, element, *args):
        del args
        context_dict = self._get_instance_dict(context)
        if element not in context_dict:
            return
        context_dict[element][6] = True

    def _tick(self, dt) -> None:
        del dt
        # run this at the end of the tick to make sure all kivy bind callbacks have executed
        Clock.schedule_once(self._render_all, -1)

    def _render_all(self, dt):
        del dt
        for context, instances in self.instances.items():
            for element, instance in instances.items():
                if not element[5] or not element[6]:
                    continue
                self._render(instance, element, context)

    # pylint: disable-msg=too-many-locals
    def _render(self, instance, element, context):
        fbo, effect_widget, source, settings, first, _, _, _ = instance
        instance[4] = False
        instance[6] = False

        # detach the widget from the parent
        parent = source.parent
        if parent and hasattr(parent, "remove_display_source"):
            parent.remove_display_source(source)

        effect_widget.add_widget(source.container)

        fbo.draw()

        fbo.bind()
        data = glReadPixels(0, 0, source.native_size[0], source.native_size[1],
                            GL_RGBA, GL_UNSIGNED_BYTE)

        fbo.release()

        effect_widget.remove_widget(source.container)

        # reattach to the parent
        if parent and hasattr(parent, "add_display_source"):
            parent.add_display_source(source)

        if not first:
            # for some reasons we got garbage in the first buffer. we just skip it for now
            values = {}
            width = source.native_size[0]
            height = source.native_size[1]
            for x, y, name in settings['light_map']:
                x_pixel = int(x * width)
                y_pixel = height - int(y * height)
                if (data[width * y_pixel * 4 + x_pixel * 4 + 3]) == 0:
                    # pixel is transparent
                    value = -1
                else:
                    value = (
                        data[width * y_pixel * 4 + x_pixel * 4],
                        data[width * y_pixel * 4 + x_pixel * 4 + 1],
                        data[width * y_pixel * 4 + x_pixel * 4 + 2])

                if name not in self._last_color or self._last_color[name] != value:
                    self._last_color[name] = value
                    values[name] = value

            self.machine.bcp_processor.send("trigger", name="display_light_player_apply", context=context,
                                            values=values, element=element, _silent=True)
        # clear the fbo background
        fbo.bind()
        fbo.clear_buffer()
        fbo.release()

    def clear_context(self, context):
        context_dict = self._get_instance_dict(context)
        for _, instance in context_dict.items():
            instance[2].canvas.remove(instance[7])
        self._reset_instance_dict(context)


mc_player_cls = McDisplayLightPlayer



from kivy.properties import OptionProperty

from mpfmc.uix.transitions import MpfTransition


class PushTransition(MpfTransition):
    """Push Transition, the incoming slide "pushes" the existing slide out
    of the frame. Can be used from any direction."""

    direction = OptionProperty('left', options=('left', 'right', 'up', 'down'))
    """String name of the direction of the transition.

    Can be 'left', 'right', 'up', or 'down'. Default is 'left'.

    """
    def on_progress(self, progression):

        s_in, s_out, width, height, progress = self.get_vars(progression)

        direction = self.direction

        if direction == 'left':
            s_in.y = s_out.y
            s_in.x = width * (1 - progress)
            s_out.x = 0 - width * progress

        elif direction == 'right':
            s_in.y = s_out.y
            s_out.x = width * progress
            s_in.x = 0 - width * (1 - progress)

        elif direction == 'down':
            s_in.x = s_out.x
            s_in.y = height * (1 - progress)
            s_out.y = 0 - height * progress

        elif direction == 'up':
            s_in.x = s_out.x
            s_out.y = height * progress
            s_in.y = 0 - height * (1 - progress)


transition_cls = PushTransition
name = 'push'



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase


class TestSound(MpfIntegrationTestCase):

    def get_machine_path(self):
        return 'integration/machine_files/sound'

    def get_config_file(self):
        return 'config.yaml'

    def test_sound_player_with_conditionals(self):
        self.machine.variables.set_machine_var("factory_reset_current_selection", 1)
        self.mock_mc_event("sounds_play")
        self.advance_time_and_run(1)
        self.assertMcEventNotCalled("sounds_play")
        self.machine.variables.set_machine_var("factory_reset_current_selection", 2)
        self.advance_time_and_run(1)
        self.assertMcEventCalled("sounds_play")



from copy import deepcopy

import queue
import logging
from distutils.version import LooseVersion

import psutil
from kivy.clock import Clock

import mpf.core.bcp.bcp_socket_client as bcp
from mpfmc._version import __bcp_version__, version as mc_version, extended_version as mc_extended_version
from mpfmc.core.bcp_server import BCPServer


class BcpProcessor:
    def __init__(self, mc):
        self.mc = mc
        self.log = logging.getLogger('BcpProcessor')

        self.socket_thread = None
        self.connected = False
        self.receive_queue = queue.Queue()
        self.sending_queue = queue.Queue()
        self.mc_process = psutil.Process()

        if self.mc.options['bcp']:
            self.mc.events.add_handler('init_done', self._start_socket_thread)
            self.enabled = True
        else:
            self.enabled = False

        self.debug_log = self.mc.machine_config['bcp']['debug']

        self.bcp_commands = {'error': self._bcp_error,
                             'goodbye': self._bcp_goodbye,
                             'hello': self._bcp_hello,
                             'machine_variable': self._bcp_machine_variable,
                             'mode_start': self._bcp_mode_start,
                             'mode_stop': self._bcp_mode_stop,
                             'mode_list': self._bcp_mode_list,
                             'ball_start': self._bcp_ball_start,
                             'ball_end': self._bcp_ball_end,
                             'player_added': self._bcp_player_add,
                             'player_turn_start': self._bcp_player_turn_start,
                             'player_variable': self._bcp_player_variable,
                             'reset': self._bcp_reset,
                             'settings': self._bcp_settings,
                             'status_request': self._bcp_status_request,
                             'switch': self._bcp_switch,
                             'trigger': self._bcp_trigger,
                             }

        self.mc.events.add_handler('client_connected', self._client_connected)
        self.mc.events.add_handler('mc_reset_complete', self._reset_complete)

        Clock.schedule_interval(self._get_from_queue, 0)

    def _client_connected(self, **kwargs):
        del kwargs
        self.send(bcp_command="set_machine_var", name="mc_version", value=mc_version)
        self.send(bcp_command="set_machine_var", name="mc_extended_version", value=mc_extended_version)
        self.send(bcp_command="monitor_start", category="machine_vars")
        self.send(bcp_command="monitor_start", category="player_vars")
        self.send(bcp_command="monitor_start", category="modes")
        self.send(bcp_command="monitor_start", category="core_events")
        self.send(bcp_command="monitor_start", category="status_request")
        self.register_trigger("master_volume_increase")
        self.register_trigger("master_volume_decrease")
        self.register_trigger("debug_dump_stats")
        self.connected = True

    def register_trigger(self, event):
        """Register a trigger for events from MPF."""
        self.send("register_trigger", event=event)

    def _start_socket_thread(self, **kwargs):
        del kwargs

        if self.socket_thread:
            return

        self.socket_thread = BCPServer(self.mc, self.receive_queue,
                                       self.sending_queue)
        self.socket_thread.daemon = True
        self.socket_thread.start()

        self.mc.events.remove_handler(self._start_socket_thread)

    def send(self, bcp_command, callback=None, rawbytes=None, **kwargs):
        """Sends a BCP command to the connected pinball controller.

        Note that if the BCP server is not running, this method will just throw
        the BCP away. (It will still call the callback in that case.

        Args:
            bcp_command: String of the BCP command name.
            callback: Optional callback method that will be called when the
                command is sent.
            **kwargs: Optional additional kwargs will be added to the BCP
                command string.

        """
        if self.enabled:
            if not self.mc.bcp_client_connected:
                raise AssertionError("Not connected to MPF.")

            self.sending_queue.put(
                (bcp.encode_command_string(bcp_command, **kwargs), rawbytes))

        if callback:
            callback()

    def receive_bcp_message(self, msg):
        """Receives an incoming BCP message to be processed.

        Note this method is intended for testing. Usually BCP messages are
        handled by the BCP Server thread, but for test purposes it's possible
        to run mpf-mc without the BCP Server, so in that case you can use this
        method to send BCP messages into the mpf-mc.

        Args:
            msg: A string of the BCP message (in the standard BCP format:
                command?param1=value1&param2=value2...

        """
        cmd, kwargs = bcp.decode_command_string(msg)
        self.receive_queue.put((cmd, kwargs))

    def _get_from_queue(self, dt):
        """Gets and processes all queued up incoming BCP commands."""
        del dt

        while not self.receive_queue.empty():
            cmd, kwargs = self.receive_queue.get(False)
            self._process_command(cmd, **kwargs)

    def _process_command(self, bcp_command, **kwargs):
        if self.debug_log:
            if 'rawbytes' in kwargs:
                debug_kwargs = deepcopy(kwargs)
                debug_kwargs['rawbytes'] = '<{} bytes>'.format(
                    len(debug_kwargs.pop('rawbytes')))

                self.log.debug("Processing command: %s %s", bcp_command,
                               debug_kwargs)
            else:
                self.log.debug("Processing command: %s %s", bcp_command,
                               kwargs)

        # Can't use try/except KeyError here because there could be a KeyError
        # in the callback which we don't want it to swallow.
        if bcp_command in self.bcp_commands:
            self.bcp_commands[bcp_command](**kwargs)
        else:
            self.log.warning("Received invalid BCP command: %s", bcp_command[:9])
            # self.send('error',message='invalid command', command=bcp_command)

    def _bcp_status_request(self, **kwargs):
        """Status request."""
        del kwargs

        self.send("status_report",
                  cpu=self.mc_process.cpu_percent(),
                  rss=self.mc_process.memory_info().rss,
                  vms=self.mc_process.memory_info().vms)

    def _bcp_hello(self, **kwargs):
        """Processes an incoming BCP 'hello' command."""
        try:
            if LooseVersion(kwargs['version']) == (
                    LooseVersion(__bcp_version__)):
                self.send('hello', version=__bcp_version__)
            else:
                self.send('hello', version='unknown protocol version')
        except KeyError:
            self.log.warning("Received invalid 'version' parameter with 'hello'")

    def _bcp_goodbye(self, **kwargs):
        """Processes an incoming BCP 'goodbye' command."""
        # if self.config['mpf-mc']['exit_on_disconnect']:
        #     self.socket_thread.sending_thread.stop()
        #     sys.exit()

    def _bcp_mode_start(self, name=None, priority=0, **kwargs):
        """Processes an incoming BCP 'mode_start' command."""
        del kwargs

        if not name:
            return
            # todo raise error

        if name == 'game':
            self.mc.game_start()

        if name in self.mc.modes:
            self.mc.modes[name].start(mode_priority=int(priority))

    def _bcp_mode_list(self, **kwargs):
        pass

    def _bcp_mode_stop(self, name, **kwargs):
        """Processes an incoming BCP 'mode_stop' command."""
        del kwargs
        if not name:
            return
            # todo raise error

        if name == 'game':
            self.mc.game_end()

        if name in self.mc.modes:
            self.mc.modes[name].stop()

    @staticmethod
    def _bcp_ball_start(player_num, ball, **kwargs):
        del player_num, ball, kwargs

    @staticmethod
    def _bcp_ball_end(**kwargs):
        del kwargs

    def _bcp_settings(self, settings, **kwargs):
        del kwargs
        for setting in settings:
            self.mc.settings.add_setting(setting)

    def _bcp_error(self, **kwargs):
        """Processes an incoming BCP 'error' command."""
        del kwargs
        self.log.warning('Received error command from client')

    def _bcp_player_add(self, player_num, **kwargs):
        """Processes an incoming BCP 'player_add' command."""
        del kwargs
        self.mc.add_player(int(player_num))

    # pylint: disable-msg=too-many-arguments
    def _bcp_player_variable(self, name, value, prev_value, change, player_num,
                             **kwargs):
        """Processes an incoming BCP 'player_variable' command."""
        del prev_value
        del change
        del kwargs
        self.mc.update_player_var(name, value, int(player_num))

    def send_machine_var_to_mpf(self, name, value):
        """Set machine var in MPF via BCP."""
        self.send("set_machine_var", name=name, value=value)

    def _bcp_machine_variable(self, name, value, change=True, prev_value=None,
                              **kwargs):
        """Processes an incoming BCP 'machine_variable' command."""
        del kwargs
        self.mc.receive_machine_var_update(name, value, change, prev_value)

    def _bcp_player_turn_start(self, player_num, **kwargs):
        """Processes an incoming BCP 'player_turn_start' command."""
        del kwargs
        self.mc.player_start_turn(int(player_num))

    def _bcp_trigger(self, name, **kwargs):
        """Processes an incoming BCP 'trigger' command."""
        self.mc.events.post(name, **kwargs)

    def _bcp_switch(self, name, state, **kwargs):
        """Processes an incoming BCP 'switch' command."""
        del kwargs
        if int(state):
            self.mc.events.post('switch_' + name + '_active')
            '''event: switch_(name)_active
            config_section: switches
            class_label: switch

            desc: Posted on MPF-MC only (e.g. not in MPF) when the MC receives
            a BCP "switch" active command. Useful for video modes and graphical
            menu navigation. Note that this is not posted for every switch all
            the time, rather, only for switches that have been configured to
            send events to BCP.
            '''
        else:
            self.mc.events.post('switch_' + name + '_inactive')
            '''event: switch_(name)_inactive
            config_section: switches
            class_label: switch

            desc: Posted on MPF-MC only (e.g. not in MPF) when the MC receives
            a BCP "switch" inactive command. Useful for video modes and graphical
            menu navigation. Note that this is not posted for every switch all
            the time, rather, only for switches that have been configured to
            send events to BCP.
            '''

    def _bcp_reset(self, **kwargs):
        del kwargs
        self.mc.reset()

    def _reset_complete(self, **kwargs):
        del kwargs
        self.send('reset_complete')



"""A slide which can show widgets."""
from bisect import bisect
from typing import List, Optional

from kivy.graphics.vertex_instructions import Rectangle
from kivy.uix.screenmanager import Screen
from kivy.uix.stencilview import StencilView
from kivy.uix.widget import Widget as KivyWidget
from kivy.graphics import Color
from kivy.properties import ListProperty, AliasProperty

from mpfmc.uix.widget import (WidgetContainer, Widget,
                              create_widget_objects_from_config)
from mpfmc.core.mc import MpfMc


# pylint: disable-msg=too-many-instance-attributes
class Slide(Screen, StencilView):

    """A slide on a display."""

    next_id = 0

    @classmethod
    def get_next_id(cls) -> int:
        """Return the next slide id."""
        Slide.next_id += 1
        return Slide.next_id

    # pylint: disable-msg=too-many-arguments
    def __init__(self, mc: "MpfMc", name: Optional[str], config: Optional[dict] = None,
                 target: str = 'default', key: Optional[str] = None,
                 priority: int = 0, play_kwargs: Optional[dict] = None) -> None:
        """Initialise slide."""
        # config is a dict. widgets will be in a key
        # assumes config, if present, is validated.
        self.creation_order = Slide.get_next_id()

        if not name:
            name = 'Anon_{}'.format(self.creation_order)

        self.mc = mc
        self.name = name
        self.priority = priority
        self.pending_widgets = set()
        self.key = key
        self.mc.track_leak_reference(self)

        if not config:
            config = self.mc.config_validator.validate_config('slides', dict())

        self.transition_out = config.get('transition_out', None)
        self.expire = config.get('expire', None)

        self.display = self.mc.targets[target]

        self.size_hint = (None, None)
        super().__init__()
        self.size = self.display.native_size
        self.orig_w, self.orig_h = self.size
        self.z = 0

        if 'widgets' in config:  # don't want try, swallows too much
            widgets = create_widget_objects_from_config(
                mc=self.mc,
                config=config['widgets'], key=self.key,
                play_kwargs=play_kwargs)

            self.add_widgets(widgets)

        self.display.add_widget(self)
        self.mc.active_slides[name] = self
        self.mc.slides[name] = config

        self.background_color = config.get('background_color', [0.0, 0.0, 0.0, 1.0])
        if self.background_color != [0.0, 0.0, 0.0, 0.0]:
            with self.canvas.before:    # noqa
                Color(*self.background_color)
                Rectangle(size=self.size, pos=(0, 0))

        self.opacity = config.get('opacity', 1.0)

        self.mc.post_mc_native_event(
            'slide_{}_created'.format(self.name))

        """event: slide_(name)_created
        config_section: slides
        class_label: slide

        desc: A slide called (name) has just been created.

        This means that this slide now exists, but it's not necessarily the
        active (showing) slide, depending on the priorities of the other slides
        and/or what else is going on.

        This is useful for things like the widget_player where you want to
        target a widget for a specific slide, but you can only do so if
        that slide exists.

        Slide names do not take into account what display or slide frame
        they're playing on, so be sure to create machine-wide unique names
        when you're naming your slides.

        """

    def __repr__(self):
        return '<Slide name={}, priority={}, id={}>'.format(self.name,
                                                            self.priority,
                                                            self.creation_order)

    def add_widgets_from_library(self, name: str, key: Optional[str] = None,
                                 widget_settings: Optional[dict] = None,
                                 play_kwargs: Optional[dict] = None,
                                 **kwargs) -> List["Widget"]:
        """Add a widget to the slide by name from the library of pre-defined widgets.

        Args:
            name: The name of the widget to add.
            key: An optional key.
            widget_settings: An optional dictionary of widget settings to override those in
                the library.
            play_kwargs: An optional dictionary of play settings to override those in
                the library.
            **kwargs:  Additional arguments.

        Returns:
            A list of widgets (MpfWidget objects) added to the slide.
        """
        del kwargs

        if name not in self.mc.widgets:
            raise ValueError("Widget {} not found".format(name))

        widgets_added = create_widget_objects_from_config(config=self.mc.widgets[name],
                                                          mc=self.mc,
                                                          key=key,
                                                          widget_settings=widget_settings,
                                                          play_kwargs=play_kwargs)
        for widget in widgets_added:
            self.add_widget(widget)

        return widgets_added

    def add_widgets(self, widgets: List["Widget"]):
        """Adds a list of widgets to this slide."""
        for w in widgets:
            self.add_widget(w)

    # pylint: disable-msg=arguments-differ
    def add_widget(self, widget: "Widget", **kwargs) -> None:
        """Add a widget to this slide.

        Args:
            widget: An MPF-enhanced widget (which will include details like z
                order and removal keys.)

        Notes:
            Widgets are drawn in order from the end of the children list to the
            beginning, meaning the first item in the child list is draw last so
            it will appear on top of all other items.

            This method respects the z-order of the widget it's adding and inserts
            it into the proper position in the widget tree. Higher numbered z order
            values will be inserted after (so they draw on top) of existing ones.

            If the new widget has the same priority of existing widgets, the new
            one is inserted after the widgets of that priority, to maintain the
            drawing order of the configuration file.
        """
        del kwargs
        if widget.get_display() == self.display:
            raise AssertionError("Cannot add widget {} to display {} because the widget uses the same display.".
                                 format(widget, self.display))

        if widget.z < 0:
            self.add_widget_to_parent_frame(widget)
            return

        # Insert the widget in the proper position in the z-order
        super().add_widget(widget, bisect(self.children, widget))

    def remove_widgets_by_key(self, key: str) -> None:
        """Removes all widgets from this slide with the specified key value."""
        for widget in self.find_widgets_by_key(key):
            if isinstance(widget, Widget):
                widget.remove()
            else:
                self.remove_widget(widget)

    def find_widgets_by_key(self, key: str) -> List["Widget"]:
        """Return a list of widgets with the matching key value by searching
        the tree of children belonging to this slide."""
        return [w for child in self.children
                for w in child.walk(restrict=True, loopback=False) if hasattr(w, "key") and w.key == key]

    def add_widget_to_parent_frame(self, widget: "KivyWidget"):
        """Adds this widget to this slide's parent instead of to this slide.

        Args:
            widget:
                The widget object.

        Notes:
            Widgets added to the parent slide_frame stay active and visible even
            if the slide in the frame changes.
        """
        # TODO: Determine proper z-order for negative z-order values
        self.manager.container.add_widget(widget)

    def schedule_removal(self, secs: float) -> None:
        """Schedules the removal of this slide after the specified number of seconds elapse."""
        self.mc.clock.schedule_once(self.remove, secs)

    def remove(self, dt=None) -> None:
        """Removes the slide from the parent display."""
        del dt

        try:
            self.manager.remove_slide(slide=self,
                                      transition_config=self.transition_out)

        except AttributeError:
            # looks like slide was already removed, but let's clean it up just
            # in case
            self.prepare_for_removal()
            self.mc.active_slides.pop(self.name, None)

    def prepare_for_removal(self) -> None:
        """Performs housekeeping chores just prior to a slide being removed."""
        self.mc.clock.unschedule(self.remove)

        for widget in self.children:
            if hasattr(widget, 'prepare_for_removal'):  # try swallows too much
                widget.prepare_for_removal()

        self.mc.post_mc_native_event('slide_{}_removed'.format(self.name))

        """event: slide_(name)_removed
        config_section: slides
        class_label: slide

        desc: A slide called (name) has just been removed.

        This event is posted whenever a slide is removed, regardless of
        whether or not that slide was active (showing).

        Note that even though this event is called "removed", it's actually
        posted as part of the removal process. (e.g. there are still some
        clean-up things that happen afterwards.)

        Slide names do not take into account what display or slide frame
        they're playing on, so be sure to create machine-wide unique names
        when you're naming your slides.

        """

    def on_pre_enter(self, *args):
        del args
        for widget in self.children:
            widget.on_pre_show_slide()

    def on_enter(self, *args):
        del args
        for widget in self.children:
            widget.on_show_slide()

    def on_pre_leave(self, *args):
        del args
        for widget in self.children:
            widget.on_pre_slide_leave()

    def on_leave(self, *args):
        del args
        for widget in self.children:
            widget.on_slide_leave()

    def on_slide_play(self):
        for widget in self.children:
            widget.on_slide_play()

    #
    # Properties
    #

    background_color = ListProperty([0.0, 0.0, 0.0, 1.0])
    '''The background color of the slide, in the (r, g, b, a)
    format.

    :attr:`background_color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [0, 0, 0, 1.0].
    '''

    def _get_parent_widgets(self) -> List["WidgetContainer"]:
        """Return the current list of widgets owned by the slide manager parent."""
        return [x for x in self.manager.parent.children if x != self.manager]

    parent_widgets = AliasProperty(_get_parent_widgets, None)
    '''List of all the :class:`MpfWidget` widgets that belong to the slide
    manager of this slide (read-only).  You should not change this list
    manually. Use the :meth:`add_widget <mpfmc.uix.widget.MpfWidget.add_widget>`
    method instead.

    Use this property rather than the 'self.manager.parent.children' property in
    case the slide architecture changes in the future.
    '''

    def _get_widgets(self) -> List["WidgetContainer"]:
        """Returns the current list of widget children owned by this slide."""
        return self.children

    widgets = AliasProperty(_get_widgets, None, bind=('children', ))
    '''List of all the :class:`MpfWidget` children widgets of this slide (read-only).
    You should not change this list manually. Use the
    :meth:`add_widget <mpfmc.uix.widget.MpfWidget.add_widget>` method instead.

    Use this property rather than the 'children' property in case the slide
    architecture changes in the future.
    '''



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc._version import version, extended_version


class TestBasicGame(MpfIntegrationTestCase):
    """Tests that MPF MC gets the basic stuff it needs from BCP from MPF.

    This includes players, games, modes, machine vars, player vars, etc.

    """

    def get_config_file(self):
        return 'basic_game.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/basic_game/'

    def test_basic_game(self):
        self.hit_switch_and_run('drain', 1)
        self.hit_and_release_switch('start')
        self.mock_event("player_added")
        self.mock_mc_event("player_added")
        self.advance_time_and_run(1)

        # make sure the MPF side has everything
        self.assertModeRunning('game')
        self.assertTrue(self.machine.game.player)
        self.assertEqual(self.machine.game.player.ball, 1)
        self.assertEqual({'value': 'FREE PLAY', 'persist': False, 'expire_secs': None},
                         self.machine.variables.machine_vars['credits_string'])

        # make sure the MC side has everything
        self.assertTrue(self.mc.modes['game'].active)
        self.assertTrue(self.mc.player)
        self.assertEqual(self.mc.player.ball, 1)
        self.assertEqual('FREE PLAY', self.mc.machine_vars['credits_string'])

        self.assertEventCalled("player_added", 1)
        self.assertMcEventCalled("player_added", 1)

        self.assertMachineVarEqual(version, "mc_version")
        self.assertMachineVarEqual(extended_version, "mc_extended_version")



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestModes(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/modes'

    def get_config_file(self):
        return 'test_modes.yaml'

    def test_mc_mode_start(self):
        # this tests that the mode is read properly from the config too
        self.send(bcp_command='mode_start', name='mode1', priority=300)
        self.assertTrue(self.mc.modes['mode1'].active)
        self.assertEqual(self.mc.modes['mode1'].priority, 300)

        # try to start the mode again and make sure it doesn't explode
        self.send(bcp_command='mode_start', name='mode1')

        # stop the mode
        self.send(bcp_command='mode_stop', name='mode1')
        self.assertFalse(self.mc.modes['mode1'].active)



""" ModeController for MPF-MC"""

import logging
import os
from collections import namedtuple

from mpf.core.config_processor import ConfigProcessor
from mpf.core.utility_functions import Util
from mpf.exceptions.config_file_error import ConfigFileError
from mpfmc.core.mode import Mode


RemoteMethod = namedtuple('RemoteMethod',
                          'method config_section kwargs priority',
                          )
"""RemotedMethod is used by other modules that want to register a method to
be called on mode_start or mode_stop.
"""

# todo create a single ModeController base class for MPF and MPF-MC


class ModeController:
    """Parent class for the Mode Controller. There is one instance of this in
    MPF and it's responsible for loading, unloading, and managing all game
    modes.
    """

    def __init__(self, mc):
        self.mc = mc
        self.log = logging.getLogger('ModeController')

        self.debug = False

        self.active_modes = list()
        self.mode_stop_count = 0

        # The following two lists hold namedtuples of any remote components
        # that need to be notified when a mode object is created and/or
        # started.
        self.loader_methods = list()
        self.start_methods = list()
        self.stop_methods = list()

        self._machine_mode_folders = dict()
        self._mpf_mode_folders = dict()

        if 'modes' in self.mc.machine_config:
            self.mc.events.add_handler('init_phase_2',
                                       self._load_modes)

    def _load_modes(self, **kwargs):
        del kwargs
        # Loads the modes from the modes: section of the machine configuration
        # file.

        self._build_mode_folder_dicts()

        for mode in set(self.mc.machine_config['modes']):
            self.mc.modes[mode] = self._load_mode(mode)

        # initialise modes after loading all of them to prevent races
        for item in self.loader_methods:
            for mode in self.mc.modes.values():
                try:
                    if ((item.config_section in mode.config and
                            mode.config[item.config_section]) or not
                            item.config_section):
                        item.method(config=mode.config.get(item.config_section),
                                    mode=mode,
                                    mode_path=mode.path,
                                    root_config_dict=mode.config,
                                    **item.kwargs)
                except ConfigFileError as e:
                    e.extend("Error while loading mode '{}'".format(mode.name))
                    raise e

    def _load_mode(self, mode_string):
        """Loads a mode, reads in its config, and creates the Mode object.

        Args:
            mode: String name of the mode you're loading. This is the name of
                the mode's folder in your game's machine_files/modes folder.

        """
        if self.debug:
            self.log.debug('Processing mode: %s', mode_string)

        asset_paths = []
        mode_path = None

        if mode_string in self._mpf_mode_folders:
            mode_path = os.path.join(
                self.mc.mpf_path,
                "modes",
                self._mpf_mode_folders[mode_string])
            asset_paths.append(mode_path)

        if mode_string in self._machine_mode_folders:
            mode_path = os.path.join(
                self.mc.machine_path,
                self.mc.machine_config['mpf-mc']['paths']['modes'],
                self._machine_mode_folders[mode_string])
            asset_paths.append(mode_path)

        if not mode_path:
            raise ValueError("No folder found for mode '{}'. Is your mode "
                             "folder in your machine's 'modes' folder?"
                             .format(mode_string))

        config = self.mc.mc_config.get_mode_config(mode_string)

        # validate config
        self.mc.config_validator.validate_config("mode", config['mode'])

        return Mode(self.mc, config, mode_string, mode_path, asset_paths)

    def _build_mode_folder_dicts(self):
        self._mpf_mode_folders = self._get_mode_folder(self.mc.mpf_path)
        self.log.debug("Found MPF Mode folders: %s", self._mpf_mode_folders)

        self._machine_mode_folders = (
            self._get_mode_folder(self.mc.machine_path))
        self.log.debug("Found Machine-specific Mode folders: %s",
                       self._machine_mode_folders)

    def _get_mode_folder(self, base_folder):
        try:
            mode_folders = os.listdir(os.path.join(
                base_folder, self.mc.machine_config['mpf-mc']['paths']['modes']))
        except FileNotFoundError:
            return dict()

        final_mode_folders = dict()

        for folder in mode_folders:

            this_mode_folder = os.path.join(
                base_folder,
                self.mc.machine_config['mpf-mc']['paths']['modes'],
                folder)

            if os.path.isdir(this_mode_folder) and not folder.startswith('_'):
                final_mode_folders[folder] = folder

        return final_mode_folders

    def register_load_method(self, load_method, config_section_name=None,
                             priority=0, **kwargs):
        """Used by system components, plugins, etc. to register themselves with
        the Mode Controller for anything they need a mode to do when it's
        registered.

        Args:
            load_method: The method that will be called when this mode code
                loads.
            config_section_name: An optional string for the section of the
                configuration file that will be passed to the load_method when
                it's called.
            priority: Int of the relative priority which allows remote methods
                to be called in a specific order. Default is 0. Higher values
                will be called first.
            **kwargs: Any additional keyword arguments specified will be passed
                to the load_method.

        Note that these methods will be called once, when the mode code is
        first
        initialized during the MPF boot process.

        """

        if not callable(load_method):
            raise ValueError("Cannot add load method '{}' as it is not"
                             "callable".format(load_method))

        self.loader_methods.append(RemoteMethod(method=load_method,
                                                config_section=config_section_name,
                                                kwargs=kwargs,
                                                priority=priority))

        self.loader_methods.sort(key=lambda x: x.priority, reverse=True)

    def register_start_method(self, start_method, config_section_name=None,
                              priority=0, **kwargs):
        """Used by system components, plugins, etc. to register themselves with
        the Mode Controller for anything that they a mode to do when it starts.

        Args:
            start_method: The method that will be called when this mode code
                loads.
            config_section_name: An optional string for the section of the
                configuration file that will be passed to the start_method when
                it's called.
            priority: Int of the relative priority which allows remote methods
                to be called in a specific order. Default is 0. Higher values
                will be called first.
            **kwargs: Any additional keyword arguments specified will be passed
                to the start_method.

        Note that these methods will be called every single time this mode is
        started.

        """

        if not callable(start_method):
            raise ValueError("Cannot add start method '{}' as it is not"
                             "callable".format(start_method))

        self.start_methods.append(RemoteMethod(method=start_method,
                                               config_section=config_section_name,
                                               priority=priority,
                                               kwargs=kwargs))

        self.start_methods.sort(key=lambda x: x.priority, reverse=True)

    def register_stop_method(self, callback, priority=0):
        # these are universal, in that they're called every time a mode stops
        # priority is the priority they're called. Has nothing to do with mode
        # priority

        if not callable(callback):
            raise ValueError("Cannot add stop method '{}' as it is not"
                             "callable".format(callback))

        self.stop_methods.append((callback, priority))

        self.stop_methods.sort(key=lambda x: x[1], reverse=True)

    def active_change(self, mode, active):
        # called when a mode goes active or inactive

        if active:
            self.active_modes.append(mode)
        else:
            self.active_modes.remove(mode)

        # sort the active mode list by priority
        self.active_modes.sort(key=lambda x: x.priority, reverse=True)



"""Contains the parent class for custom code used in MPF-MC."""
from typing import Callable
from mpf.core.logging import LogMixin

MYPY = False
if MYPY:    # pragma: no cover
    from mpfmc.core.mc import MpfMc


class McCustomCode(LogMixin):

    """Custom code in MPF-MC."""

    def __init__(self, mc, name):
        super().__init__()
        self.mc = mc        # type: MpfMc
        self.name = name    # type: str
        self.configure_logging('McCustomCode.' + name, 'basic', 'full')
        self.on_load()
        # wait until MC connected to MPF
        self.mc.events.add_handler("client_connected", self.on_connect)

    def add_mpf_event_handler(self, event: str, handler: Callable):
        """Add an event handler to listen to a MPF event."""
        self.mc.bcp_processor.register_trigger(event)
        return self.mc.events.add_handler(event, handler)

    def post_event_to_mpf_and_mc(self, event, **kwarg):
        """Post an event to MPF and MC."""
        self.mc.post_mc_native_event(event, **kwarg)

    def __repr__(self):  # pragma: no cover
        return '<CustomCode.{}>'.format(self.name)

    def on_connect(self, **kwargs):
        """Called once MC is connected to MPF."""

    def on_load(self):  # pragma: no cover
        """Automatically called when this custom code class loads.

        It's the intention that the custom code writer will overwrite this method.
        """



# pylint: disable-msg=too-many-lines
"""A widget on a slide."""
from typing import Union, Optional, List, Tuple
from copy import deepcopy
from functools import reduce
import math

from kivy.clock import Clock
from kivy.animation import Animation
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.widget import Widget as KivyWidget
from kivy.properties import (NumericProperty, ReferenceListProperty,
                             StringProperty, AliasProperty, ListProperty)

from mpf.core.rgba_color import RGBAColor

from mpfmc.uix.relative_animation import RelativeAnimation
from mpfmc.core.utils import percent_to_float
from mpfmc.uix.widget_magic_events import magic_events

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


# pylint: disable-msg=too-many-instance-attributes
class Widget(KivyWidget):
    """MPF-MC Widget class.

    The :class:`Widget` class is the base class required for creating Widgets
    for use in the media controller.  It is based on the Kivy
    kivy.uix.widget.Widget class, but has some custom behavior for use in
    the MC.

    The most important detail is every widget is contained inside another
    specialized widget class (mpfmc.uix.widget.WidgetContainer).  This
    container class is always the parent of a MC Widget and provides the
    coordinate translations to allow MC widgets to use their anchor point
    coordinates instead of the bottom-left corner for all coordinate settings
    (x, y, pos).  The WidgetContainer is automatically created when a widget
    is created and should not be manipulated directly.  It is important to
    remember when walking the widget tree the WidgetContainer is the Widget's
    parent.

    """

    widget_type_name = ''  # Give this a name in your subclass, e.g. 'Image'

    # We loop through the keys in a widget's config dict and check to see if
    # the widget's base class has attributes for them, and if so, we set
    # them. This is how any attribute from the base class can be exposed via
    # our configs. However we use some config keys that Kivy also uses,
    # and we use them for different purposes, so there are some keys that we
    # use that we never want to set on widget base classes.
    _dont_send_to_kivy = ('x', 'y', 'key')

    merge_settings = tuple()

    animation_properties = list()
    """List of properties for this widget that may be animated using widget animations."""

    def __init__(self, mc: "MpfMc", config: Optional[dict] = None,
                 key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        self._container = None
        self.size_hint = (None, None)

        # Needs to be deepcopy since configs can have nested dicts
        self.config = deepcopy(config)

        super().__init__(**self.pass_to_kivy_widget_init())

        self.mc = mc
        self.mc.track_leak_reference(self)

        self.animation = None
        self._animation_event_keys = set()
        # MPF event keys for event handlers that have been registered for
        # animation events. Used to remove the handlers when this widget is
        # removed.

        self._pre_animated_settings = dict()
        # dict of original values of settings that were animated so we can
        # restore them later

        self._percent_prop_dicts = dict()

        self._round_anchor_styles = (None, None)

        self._default_style = None
        self._set_default_style()
        self._apply_style()

        # Create a container widget as this widget's parent.  The container will adjust
        # the coordinate system for this widget so that all positional properties are
        # based on the widget's anchor rather than the lower left corner.
        self._container = WidgetContainer(self, z=self.config['z'])
        self._container.add_widget(self)
        self._container.fbind('parent', self.on_container_parent)

        if 'color' in self.config and not isinstance(self.config['color'], RGBAColor):
            self.config['color'] = RGBAColor(self.config['color'])

        # Set initial attribute values from config
        for k, v in self.config.items():
            if k not in self._dont_send_to_kivy and hasattr(self, k):
                setattr(self, k, v)

        # Has to be after we set the attributes since it could be in the config
        self.key = key

        # Build animations
        if 'animations' in self.config and self.config['animations']:
            for k, v in self.config['animations'].items():
                if k.split("{")[0] == 'add_to_slide':
                    # needed because the initial properties of the widget
                    # aren't set yet
                    Clock.schedule_once(self.on_add_to_slide, -1)

                elif k not in magic_events:
                    self._register_animation_events(k)
        else:
            self.config['animations'] = dict()

        # why is this needed? Why is it not config validated by here? todo
        if 'reset_animations_events' in self.config:
            for event in [x for x in self.config['reset_animations_events'] if x not in magic_events]:
                self._animation_event_keys.add(self.mc.events.add_handler(
                    event=event, handler=self.reset_animations))

        # Set widget expiration (if configured)
        self.expire = config.get('expire', None)
        if self.expire:
            self.schedule_removal(self.expire)

        # Send custom user events when widget is added (if any are configured)
        if self.config['events_when_added'] is not None:
            for event in self.config['events_when_added']:
                self.mc.post_mc_native_event(event)

    def __repr__(self) -> str:  # pragma: no cover
        return '<{} Widget id={}>'.format(self.widget_type_name, id(self))

    @staticmethod
    def get_display():
        """Get the display used"""
        return None

    @staticmethod
    def pass_to_kivy_widget_init() -> dict:
        """Initializes the dictionary of settings to pass to Kivy."""
        return dict()

    def merge_asset_config(self, asset) -> None:
        for setting in [x for x in self.merge_settings if (
                        x not in self.config['_default_settings'] and
                        x in asset.config)]:
            self.config[setting] = asset.config[setting]

    def on_anchor_offset_pos(self, instance, pos):
        """Called whenever the anchor_offset_pos property value changes."""
        del instance
        if self.parent:
            self.parent.pos = pos

    def on_container_parent(self, instance, parent):
        del instance
        if parent:
            # some attributes can be expressed in percentages. This dict holds
            # those, key is attribute name, val is max value

            self._percent_prop_dicts = dict(x=parent.width,
                                            y=parent.height,
                                            width=parent.width,
                                            height=parent.height,
                                            opacity=1,
                                            line_height=1)

            # The top-most parent owns the display, so traverse up to find the config
            top_widget = parent
            while not hasattr(top_widget, "display") and top_widget.parent != top_widget and top_widget.parent:
                print(top_widget)
                top_widget = top_widget.parent
            displayconfig = top_widget.display.config if hasattr(top_widget, 'display') else dict()

            # If the positioning is centered, look for a rounding setting to avoid
            # fractional anchor positions. Fallback to display's config if available
            round_anchor_x = self.config['round_anchor_x'] or displayconfig.get('round_anchor_x')
            round_anchor_y = self.config['round_anchor_y'] or displayconfig.get('round_anchor_y')

            # Store the anchor rounding config from widget/display to avoid recalculation
            self._round_anchor_styles = (round_anchor_x, round_anchor_y)

            self.pos = self.calculate_initial_position(parent.width,
                                                       parent.height,
                                                       self.config['x'],
                                                       self.config['y'],
                                                       round_anchor_x,
                                                       round_anchor_y)

            # Update the initial widget position based on the rounding config
            self.pos = self.calculate_rounded_position(self.anchor_offset_pos)

    def calculate_rounded_position(self, anchor: Tuple[int, int]) -> tuple:
        """Returns a tuple of (x, y) coordinates for the position of the widget,
        accounting for odd-numbered pixel dimensions and the rounding configuration
        of the widget/display."""
        # Start with the given initial position of the widget
        rounded_x = self.pos[0]
        rounded_y = self.pos[1]

        # Shift each of the x/y coordinates according to the anchor rounding
        if self._round_anchor_styles[0] == 'left':
            rounded_x -= anchor[0] % 1
        elif self._round_anchor_styles[0] == 'right':
            rounded_x += (1 - anchor[0]) % 1
        if self._round_anchor_styles[1] == 'bottom':
            rounded_y -= anchor[1] % 1
        elif self._round_anchor_styles[1] == 'top':
            rounded_y += (1 - anchor[1]) % 1

        return rounded_x, rounded_y

    @staticmethod
    def _calculate_x_position(parent_w: int, x: Optional[Union[int, str]] = None,
                              round_x: Optional[Union[bool, str]] = None) -> float:
        # ----------------------
        # X / width / horizontal
        # ----------------------
        if x is None:
            x = 'center'
        # Calculate position
        if isinstance(x, str):

            x = str(x).replace(' ', '')
            start_x = 0

            if x.startswith('right'):
                x = x.strip('right')
                start_x = parent_w

            elif x.startswith('middle'):
                x = x.strip('middle')
                start_x = parent_w / 2

            elif x.startswith('center'):
                x = x.strip('center')
                start_x = parent_w / 2

            elif x.startswith('left'):
                x = x.strip('left')

            if not x:
                x = '0'

            x = percent_to_float(x, parent_w)
            x += start_x

            if round_x == 'left':
                x = math.floor(x)
            elif round_x == 'right':
                x = math.ceil(x)

        return x

    @staticmethod
    def _calculate_y_position(parent_h: int, y: Optional[Union[int, str]] = None,
                              round_y: Optional[Union[bool, str]] = None) -> float:
        # Set defaults
        if y is None:
            y = 'middle'

        # --------------------
        # Y / height / vertical
        # --------------------

        # Calculate position
        if isinstance(y, str):

            y = str(y).replace(' ', '')
            start_y = 0

            if y.startswith('top'):
                y = y.strip('top')
                start_y = parent_h

            elif y.startswith('middle'):
                y = y.strip('middle')
                start_y = parent_h / 2

            elif y.startswith('center'):
                y = y.strip('center')
                start_y = parent_h / 2

            elif y.startswith('bottom'):
                y = y.strip('bottom')

            if not y:
                y = '0'

            y = percent_to_float(y, parent_h)
            y += start_y

            if round_y == 'bottom':
                y = math.floor(y)
            elif round_y == 'top':
                y = math.ceil(y)

        return y

    # pylint: disable-msg=too-many-arguments
    @classmethod
    def calculate_initial_position(cls, parent_w: int, parent_h: int,
                                   x: Optional[Union[int, str]] = None,
                                   y: Optional[Union[int, str]] = None,
                                   round_x: Optional[Union[bool, str]] = None,
                                   round_y: Optional[Union[bool, str]] = None) -> Tuple[float, float]:
        """Returns the initial x,y position for the widget within a larger
        parent frame based on several positioning parameters. This position will
        be combined with the widget anchor position to determine its actual
        position on the screen.

        Args:
            parent_w: Width of the parent frame.
            parent_h: Height of the parent frame.
            x: (Optional) Specifies the x (horizontal) position of the widget from
                the left edge of the slide. Can be a numeric value which
                represents the actual x value, or can be a percentage (string with
                percent sign, like '20%') which is set taking into account the size
                of the parent width. (e.g. parent width of 800 with x='20%'
                results in x=160. Can also be negative to position the widget
                partially off the left of the slide. Default value of None will
                return the horizontal center (parent width / 2). Can also start
                with the strings "left", "center", or "right" which can be combined
                with values. (e.g right-2, left+4, center-1)
            y: (Optional) Specifies the y (vertical) position of the widget from
                the bottom edge of the slide. Can be a numeric value which
                represents the actual y value, or can be a percentage (string with
                percent sign, like '20%') which is set taking into account the size
                of the parent height. (e.g. parent height of 600 with y='20%'
                results in y=120. Can also be negative to position the widget
                partially off the bottom of the slide. Default value of None will
                return the vertical center (parent height / 2). Can also start
                with the strings "top", "middle", or "bottom" which can be combined
                with values. (e.g top-2, bottom+4, middle-1)
            round_x: (Optional) Specifies a direction of either "left" or "right"
                to round the calculated pixel value for the horizontal position.
                Used to prevent partial pixel placement on DMDs, especially when
                position/anchors are specified in percentages
            round_y: (Optional) Specifies a direction of either "bottom" or "top"
                to round the calculated pixel value for the vertical position.
                Used to prevent partial pixel placement on DMDs, especially when
                position/anchors are specified in percentages

        Returns: Tuple of x, y coordinates for the lower-left corner of the
            widget you're placing.

        See the widgets documentation for examples.

        """
        return cls._calculate_x_position(parent_w, x, round_x), cls._calculate_y_position(parent_h, y, round_y)

    def _set_default_style(self) -> None:
        """Sets the default widget style name."""
        if ('{}_default'.format(self.widget_type_name.lower()) in
                self.mc.machine_config['widget_styles']):
            self._default_style = self.mc.machine_config['widget_styles'][
                '{}_default'.format(self.widget_type_name.lower())]

    def _apply_style(self, force_default: bool = False) -> None:
        """Apply any style to the widget that is specified in the config."""
        if not self.config['style'] or force_default:
            if self._default_style:
                styles = [self._default_style]
            else:
                return
        else:
            try:
                styles = [self.mc.machine_config['widget_styles'][s] for s in self.config['style']]
            except KeyError as e:
                # TOOD: After sufficient time post-0.51, remove this breaking-change-related message
                if " ".join(self.config['style']) in self.mc.machine_config['widget_styles']:
                    raise ValueError("{} has an invalid style name: {}. ".format(self, e) +
                                     "Please note that as of MPF 0.51, spaces are no longer valid " +
                                     "in widget style names (see '{}')".format(" ".join(self.config['style'])))
                raise ValueError("{} has an invalid style name: {}".format(
                    self, e))

        found = False

        try:
            # This looks crazy but it's not too bad... The list comprehension
            # builds a list of attributes (settings) that are in the style
            # definition but that were not manually set in the widget.

            # Then it sets the attributes directly since the config was already
            # processed.
            for style in styles:
                for attr in [x for x in style if
                             x not in self.config['_default_settings']]:
                    self.config[attr] = style[attr]

            found = True

        except (AttributeError, KeyError):
            pass

        if not found and not force_default:
            self._apply_style(force_default=True)

    def prepare_for_removal(self) -> None:
        """Prepare the widget to be removed."""
        self.mc.clock.unschedule(self.remove)
        self.stop_animation()
        self._remove_animation_events()

        # Send custom user events when widget is removed (if any are configured)
        if self.config['events_when_removed'] is not None:
            for event in self.config['events_when_removed']:
                self.mc.post_mc_native_event(event)

    def schedule_removal(self, secs: float) -> None:
        """Schedule the widget to be removed after the specified number
        of seconds have elapsed."""
        self.mc.clock.schedule_once(self.remove, secs)

    def remove(self, *dt) -> None:
        """Perform the actual removal of the widget."""
        del dt

        self.prepare_for_removal()

        try:
            # This widget has a container parent that must be removed
            self._container.parent.remove_widget(self._container)
        except AttributeError:
            pass

        self.on_remove_from_slide()

    def _convert_animation_value_to_float(self, prop: str,
                                          val: Union[str, int, float], event_args) -> Union[float, int]:
        """
        Convert an animation property value to a numeric value.
        Args:
            prop: The name of the property to animate
            val: The animation target value (may be a string that contains a % sign)

        Returns:
            Numeric value (float or int).
        """
        if val.startswith("(") and val.endswith(")"):
            try:
                val = event_args[val[1:-1]]
            except KeyError:
                raise AssertionError("Excepted an event parameter {}".format(val[1:-1]))

        try:
            val = percent_to_float(val, self._percent_prop_dicts[prop])
        except KeyError:
            # because widget properties can include a % sign, they are
            # all strings, so even ones that aren't on the list to look
            # for percent signs have to be converted to numbers.
            if '.' in val:
                val = float(val)
            else:
                val = int(val)

        return val

    def _resolve_named_animations(self, config_list):
        # find any named animations and replace them with the real ones
        animation_list = list()

        for entry in config_list:
            if 'named_animation' in entry:
                for named_anim_settings in (
                        self.mc.animations[entry['named_animation']]):
                    animation_list.append(named_anim_settings)
            else:
                animation_list.append(entry)

        return animation_list

    # pylint: disable-msg=too-many-branches
    # pylint: disable-msg=too-many-locals
    def build_animation_from_config(self, config_list: list, event_args) -> Animation:
        """Build animation object from config."""
        if not isinstance(config_list, list):
            raise TypeError('build_animation_from_config requires a list')

        # find any named animations and replace them with the real ones
        animation_list = self._resolve_named_animations(config_list)

        repeat = False
        animation_sequence_list = []

        for settings in animation_list:
            prop_dict = dict()
            values_needed = dict()
            values = settings['value'].copy()

            # Some properties that can be animated contain more than single values
            # (such as color). Need to ensure there are the correct number of
            # values for the properties to animate.
            values_needed_total = 0

            for prop in settings['property']:
                if isinstance(getattr(self, prop), list):
                    values_needed[prop] = len(getattr(self, prop))
                    values_needed_total += values_needed[prop]
                else:
                    values_needed[prop] = 1
                    values_needed_total += 1

            if len(settings['value']) != values_needed_total:
                self.mc.log.warning("There is a mismatch between the number of values "
                                    "available and the number of values required to animate "
                                    "the following properties in the %s widget: %s "
                                    "(animation will be ignored).",
                                    self.widget_type_name, settings['property'])
                continue

            # Create a dictionary of properties to animate along with their target values
            for prop in settings['property']:

                # Make sure target widget property can be animated
                if prop not in self.animation_properties:
                    self.mc.log.warning("%s widgets do not support animation "
                                        "for the %s property (animation will be ignored)",
                                        self.widget_type_name, prop)
                    continue

                # Convert target value(s) to numeric types
                if values_needed[prop] > 1:
                    val = [self._convert_animation_value_to_float(prop, x, event_args)
                           for x in values[:values_needed[prop]]]
                    del values[:values_needed[prop]]
                else:
                    val = self._convert_animation_value_to_float(prop, values[0], event_args)
                    del values[0]

                prop_dict[prop] = val

                # Save the pre-animated property value so it can later be restored
                if prop not in self._pre_animated_settings:
                    self._pre_animated_settings[prop] = getattr(self, prop)

            # TODO: Support custom easing functions
            # This can be done by replacing transition string with a function reference
            # when the string does not exist in the Kivy AnimationTransition class as
            # a method.

            # Create the animation object
            if settings['relative']:
                animation = RelativeAnimation(duration=settings['duration'],
                                              transition=settings['easing'],
                                              **prop_dict)
            else:
                animation = Animation(duration=settings['duration'],
                                      transition=settings['easing'],
                                      **prop_dict)

            # Determine if this animation should be performed in sequence or in parallel
            # with the previous animation.
            if settings['timing'] == 'with_previous' and animation_sequence_list:
                # Combine in parallel with previous animation
                animation_sequence_list[-1] &= animation
            else:
                # Add new sequential animation to the list
                animation_sequence_list.append(animation)

            if settings['repeat']:
                repeat = True

        # Combine all animations that should be performed in sequence into a single
        # animation object (add them all together)
        final_animation = reduce(lambda x, y: x + y, animation_sequence_list)

        if repeat:
            final_animation.repeat = True

        return final_animation

    def stop_animation(self) -> None:
        """Stop the current widget animation."""
        try:
            self.animation.cancel(self)
        except AttributeError:
            pass

    def reset_animations(self, **kwargs) -> None:
        """Reset the widget properties back to their pre-animated values."""
        del kwargs
        for k, v in self._pre_animated_settings.items():
            setattr(self, k, v)

    def _register_animation_events(self, event_name: str) -> None:
        """Register handlers for the various events that trigger animation actions."""
        self._animation_event_keys.add(self.mc.events.add_handler(
            event=event_name, handler=self.start_animation_from_event,
            event_name=event_name))

    def start_animation_from_event(self, event_name: str, **kwargs) -> None:
        """Starts an animation based on an event name that has previously
        been registered."""
        if event_name not in self.config['animations']:
            return

        self.stop_animation()
        self.animation = self.build_animation_from_config(
            self.config['animations'][event_name], kwargs)
        self.animation.start(self)

    def _remove_animation_events(self) -> None:
        """Remove previously registered handlers for the various events that trigger animation actions."""
        self.mc.events.remove_handlers_by_keys(self._animation_event_keys)
        self._animation_event_keys = set()

    def on_add_to_slide(self, dt) -> None:
        """Automatically called when this widget is added to a slide.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        del dt

        if 'add_to_slide' in self.config['reset_animations_events']:
            self.reset_animations()

        for k in self.config['animations'].keys():
            event, placeholder, _ = self.mc.events.get_event_and_condition_from_string(k)
            if event == 'add_to_slide' and (not placeholder or placeholder.evaluate({})):
                self.start_animation_from_event(k)
                break

    def on_remove_from_slide(self) -> None:
        """Automatically called when this widget is removed from a slide.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'remove_from_slide' in self.config['reset_animations_events']:
            self.reset_animations()

    def on_pre_show_slide(self) -> None:
        """Automatically called when the slide this widget is part of is about
        to be shown. If there's an entrance transition, this method is called
        before the transition starts.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'pre_show_slide' in self.config['reset_animations_events']:
            self.reset_animations()

        if 'pre_show_slide' in self.config['animations']:
            self.start_animation_from_event('pre_show_slide')

    def on_show_slide(self) -> None:
        """Automatically called when the slide this widget is part of has been
        shown. If there's an entrance transition, this method is called
        after the transition is complete.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'show_slide' in self.config['reset_animations_events']:
            self.reset_animations()

        if 'show_slide' in self.config['animations']:
            self.start_animation_from_event('show_slide')

    def on_pre_slide_leave(self) -> None:
        """Automatically called when the slide this widget is part of is about
        to leave (e.g. when another slide is going to replace it). If
        there's an exit transition, this method is called before the
        transition starts.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'pre_slide_leave' in self.config['reset_animations_events']:
            self.reset_animations()

        if 'pre_slide_leave' in self.config['animations']:
            self.start_animation_from_event('pre_slide_leave')

    def on_slide_leave(self) -> None:
        """Automatically called when the slide this widget is part of is about
        to leave (e.g. when another slide is going to replace it). If there's
        an exit transition, this method is called after the transition is
        complete.

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'slide_leave' in self.config['reset_animations_events']:
            self.reset_animations()

        if 'slide_leave' in self.config['animations']:
            self.start_animation_from_event('slide_leave')

    def on_slide_play(self) -> None:
        """Automatically called when the slide this widget is part of is played
        as part of a slide_player play command (either via a standalone slide
        player or as a show step).

        If you subclass this method, be sure to call super(), as it's needed
        for widget animations.
        """
        if 'slide_play' in self.config['reset_animations_events']:
            self.reset_animations()

        if 'slide_play' in self.config['animations']:
            self.start_animation_from_event('slide_play')

    def find_widgets_by_key(self, key: str) -> List["KivyWidget"]:
        """Return a list of widgets with the matching key value by searching
        the tree of children belonging to this widget."""
        return [x for x in self.walk(restrict=True, loopback=False) if hasattr(x, 'key') and x.key == key]

    #
    # Properties
    #

    def _get_container(self) -> KivyWidget:
        return self._container

    container = AliasProperty(_get_container, None)
    '''The widget container is a special container/parent widget that manages this widget.
    It has no graphical representation.'''

    key = StringProperty(None, allownone=True)
    '''Widget keys are used to uniquely identify instances of widgets which you can later
    use to update or remove the widget.
    '''

    color = ListProperty([1.0, 1.0, 1.0, 1.0])
    '''The color of the widget, in the (r, g, b, a) format.

    :attr:`color` is a :class:`~kivy.properties.ListProperty` and
    defaults to [1.0, 1.0, 1.0, 1.0].
    '''

    anchor_x = StringProperty(None, allownone=True)
    '''Which edge of the widget will be used for positioning. ('left', 'center'
    (or 'middle'), or 'right'. If None, 'center' will be used.
    '''

    anchor_y = StringProperty(None, allownone=True)
    '''Which edge of the widget will be used for positioning. ('top', 'middle'
    (or 'center'), or 'bottom'. If None, 'center' will be used.
    '''

    anchor_pos = ReferenceListProperty(anchor_x, anchor_y)
    '''Which point of the widget will be used for positioning.

    :attr:`anchor_pos` is a :class:`~kivy.properties.ReferenceListProperty`
    of (:attr:`anchor_x`, :attr:`anchor_y`) properties.
    '''

    adjust_top = NumericProperty(0)
    '''Moves the "top" of this widget's anchor position down, meaning any
    positioning that includes calculations involving the top (anchor_y of 'top'
    or 'middle') use the alternate top position. Positive values move the top
    towards the center of the widget, negative values move it away. Negative
    values can be used to give the widget "space" on the top, and positive
    values can be used to remove unwanted space from the top of the widget.
    Note that this setting does not actually crop or cut off the top of the
    widget, rather, it just adjusts how the positioning is calculated.
    '''

    adjust_right = NumericProperty(0)
    '''Adjusts the anchor position calculations for the right side of the widget.
    Positive values move the right position  towards the center, negative values
    move it away from the center.
    '''

    adjust_bottom = NumericProperty(0)
    '''Adjusts the anchor position calculations for the bottom of the widget.
    Positive values move the bottom position towards the center, negative values
    move it away from the center.
    '''

    adjust_left = NumericProperty(0)
    '''Adjusts the anchor position calculations for the left side of the widget.
    Positive values move the left position towards the center, negative values
    move it away from the center.
    '''

    def _get_anchor_offset_pos(self):
        """Calculate the anchor offset position relative to the lower-left corner of the widget.

        Based on several positioning parameters.
        """
        # Set defaults
        offset_x = 0
        offset_y = 0
        anchor_x = self.anchor_x
        anchor_y = self.anchor_y
        if not anchor_x:
            anchor_x = 'center'
        if not self.anchor_y:
            anchor_y = 'middle'

        # ----------------------
        # X / width / horizontal
        # ----------------------

        # Adjust for anchor_x & adjust_right/left
        if anchor_x in ('center', 'middle'):
            offset_x -= (self.width - self.adjust_right + self.adjust_left) / 2
        elif anchor_x == 'right':
            offset_x -= self.width - self.adjust_right
        else:  # left
            offset_x -= self.adjust_left

        # --------------------
        # Y / height / vertical
        # --------------------

        # Adjust for anchor_y & adjust_top/bottom
        if anchor_y in ('middle', 'center'):
            offset_y -= (self.height - self.adjust_top + self.adjust_bottom) / 2
        elif anchor_y == 'top':
            offset_y -= self.height - self.adjust_top
        else:  # bottom
            offset_y -= self.adjust_bottom

        return offset_x, offset_y

    anchor_offset_pos = AliasProperty(_get_anchor_offset_pos, None,
                                      bind=('size', 'anchor_x', 'anchor_y', 'adjust_top',
                                            'adjust_right', 'adjust_bottom', 'adjust_left'),
                                      cache=True)
    '''The anchor position of the widget (relative to the widget's lower left corner).

    :attr:`anchor_offset_pos` is a :class:`~kivy.properties.ReferenceListProperty`
    of (:attr:`anchor_offset_x`, :attr:`anchor_offset_y`) properties.
    '''


def create_widget_objects_from_config(mc: "MpfMc", config: Union[dict, list],
                                      key: Optional[str] = None,
                                      play_kwargs: Optional[dict] = None,
                                      widget_settings: Optional[dict] = None) -> List["WidgetContainer"]:
    """
    Creates one or more widgets from config settings.

    Args:
        mc: An instance of MC
        config: The configuration dictionary for the widgets to be created.
        key: An optional key.
        play_kwargs: An optional dictionary of play settings to override those in
            the config.
        widget_settings: An optional dictionary of widget settings to override those in
            the config.

    Returns:
        A list of the WidgetContainer objects created.
    """
    if not isinstance(config, list):
        config = [config]
    widgets_added = list()

    if not play_kwargs:
        play_kwargs = dict()  # todo

    for widget in config:
        # Lookup a pre-defined widget based on a name
        name = widget.get('widget', None)
        if name:
            widgets_added += create_widget_objects_from_library(name=name,
                                                                mc=mc,
                                                                play_kwargs=play_kwargs,
                                                                widget_settings=widget_settings)
            continue
        if widget_settings:
            widget_settings = mc.config_validator.validate_config(
                'widgets:{}'.format(widget['type']), widget_settings,
                base_spec='widgets:common', add_missing_keys=False)

            widget.update(widget_settings)

        configured_key = widget.get('key', None)

        if configured_key and key and "." not in key and configured_key != key:
            raise KeyError("Widget has incoming key '{}' which does not "
                           "match the key in the widget's config "
                           "'{}'.".format(key, configured_key))

        if configured_key:
            this_key = configured_key
        else:
            this_key = key

        widget_obj = mc.widgets.type_map[widget['type']](
            mc=mc, config=widget, key=this_key, play_kwargs=play_kwargs)

        top_widget = widget_obj

        # some widgets have parents, so we need to make sure that we add
        # the parent widget to the slide
        while top_widget.parent:
            top_widget = top_widget.parent

        widgets_added.append(top_widget)

    return widgets_added


def create_widget_objects_from_library(mc: "MpfMc", name: str,
                                       key: Optional[str] = None,
                                       widget_settings: Optional[dict] = None,
                                       play_kwargs: Optional[dict] = None,
                                       **kwargs) -> List["WidgetContainer"]:
    """

    Args:
        mc:
        name:
        key:
        widget_settings:
        play_kwargs:
        **kwargs:

    Returns:
        A list of the MpfWidget objects created.
    """
    del kwargs

    # If the name is a placeholder template, evaluate it against the args
    if name and hasattr(name, "evaluate"):
        name = name.evaluate(play_kwargs)

    if name not in mc.widgets:
        raise ValueError("Widget {} not found".format(name))

    return create_widget_objects_from_config(mc=mc,
                                             config=mc.widgets[name],
                                             key=key,
                                             widget_settings=widget_settings,
                                             play_kwargs=play_kwargs)


class WidgetContainer(RelativeLayout):

    def __init__(self, widget: "Widget",
                 key: Optional[str] = None, z: int = 0, **kwargs) -> None:
        del kwargs
        self.key = None
        super().__init__(size_hint=(1, 1))

        self.key = key
        self.z = z
        self._widget = widget

    def __repr__(self) -> str:  # pragma: no cover
        return '<WidgetContainer z={} key={}>'.format(self.z, self.key)

    def __lt__(self, other: "KivyWidget") -> bool:
        """Less than comparison operator (based on z-order value).

        Used to maintain proper z-order when adding widgets to a parent.
        Args:
            other: The widget to compare to this one.

        Returns:
            True if the other widget is less than the current widget (uses
            z-order to perform the comparison).
        """
        if hasattr(other, 'z'):
            return other.z < self.z
        else:
            return self.z > 0

    def prepare_for_removal(self) -> None:
        if self._widget:
            self._widget.prepare_for_removal()

    def on_pre_show_slide(self) -> None:
        if self._widget:
            self._widget.on_pre_show_slide()

    def on_show_slide(self) -> None:
        if self._widget:
            self._widget.on_show_slide()

    def on_pre_slide_leave(self) -> None:
        if self._widget:
            self._widget.on_pre_slide_leave()

    def on_slide_leave(self) -> None:
        if self._widget:
            self._widget.on_slide_leave()

    def on_slide_play(self) -> None:
        if self._widget:
            self._widget.on_slide_play()

    #
    # Properties
    #

    z = NumericProperty(0)
    '''Z position (z-order) of the widget (used to determine the drawing order of widgets).
    '''

    def get_display(self):
        """Get the display used"""
        return self._widget.get_display()

    def _get_widget(self):
        return self._widget

    widget = AliasProperty(_get_widget, None)
    '''The MC Widget child of this container widget.'''



from typing import List

from kivy.properties import NumericProperty, ListProperty, BooleanProperty

from mpfmc.uix.effects import EffectsChain
from mpfmc.effects.dot_filter import DotFilterEffect
from mpfmc.effects.gain import GainEffect
from mpfmc.effects.reduce import ReduceEffect

MYPY = False
if MYPY:   # pragma: no cover
    from kivy.uix.effectwidget import EffectBase


class ColorDmdEffect(EffectsChain):
    """GLSL effect to render an on-screen DMD to look like individual round pixels."""

    dot_filter = BooleanProperty(True)
    '''
    Sets whether or not to apply the dot filter effect.

    dot_filter is a :class:`~kivy.properties.BooleanProperty` and
    defaults to True.
    '''

    width = NumericProperty(128)
    '''
    Sets the width in pixels of the display widget on the screen. Typically
    this is larger than the dots_x parameter.

    width is a :class:`~kivy.properties.NumericProperty` and
    defaults to 128.
    '''

    height = NumericProperty(32)
    '''
    Sets the height in pixels of the display widget on the screen. Typically
    this is larger than the dots_y parameter.

    height is a :class:`~kivy.properties.NumericProperty` and
    defaults to 32.
    '''

    dots_x = NumericProperty(128)
    '''
    Sets the number of dots in the horizontal (x) direction.

    dots_x is a :class:`~kivy.properties.NumericProperty` and
    defaults to 128.
    '''

    dots_y = NumericProperty(32)
    '''
    Sets the number of dots in the vertical (y) direction.

    dots_y is a :class:`~kivy.properties.NumericProperty` and
    defaults to 32.
    '''

    blur = NumericProperty(0.1)
    '''
    Sets the size of the blur around each pixel where it's blended with
    the background. The value is relative to the pixel. (e.g. a value of
    0.1 will add a 10% blur around the edge of each pixel.)

    blur is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.1.
    '''

    dot_size = NumericProperty(0.5)
    '''
    Sets the size of the circle for the dot/pixel relative to the size of the
    square bounding box of the dot. A size of 1.0 means that the diameter
    of the dot will be the same as its bounding box, in other words a
    size of 1.0 means that the dot will touch each other.

    dot_size is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.5.
    '''

    background_color = ListProperty([0.1, 0.1, 0.1, 1.0])
    '''
    A four-item tuple or list that represents the color of the space between the
    dots, in RGBA format with individual values as floats between 0.0 - 1.0. If
    you want the background to be transparent, set it to (0.0, 0.0, 0.0, 0.0).

    background_color is a :class:`~kivy.properties.ListProperty` and
    defaults to (0.1, 0.1, 0.1, 1.0) (which is 10% gray with 100% alpha/fully
    opaque).
    '''

    gain = NumericProperty(1.0)
    '''
    Sets the gain factor which is multiplied by each color channel.

    gain is a :class:`~kivy.properties.NumericProperty` and
    defaults to 1.0 (which has no effect).
    '''

    shades = NumericProperty(16)
    '''
    Sets the number of shades per channel to reduce it to.

    shades is a :class:`~kivy.properties.NumericProperty` and
    defaults to 16.
    '''

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_effects(self) -> List["EffectBase"]:
        effects = []

        if bool(self.dot_filter):
            effects.append(DotFilterEffect(
                width=self.width,
                height=self.height,
                dots_x=self.dots_x,
                dots_y=self.dots_y,
                blur=self.blur,
                dot_size=self.dot_size,
                background_color=self.background_color
            ))

        if self.shades > 0:
            effects.append(ReduceEffect(shades=self.shades))

        effects.append(GainEffect(gain=self.gain))

        return effects


effect_cls = ColorDmdEffect
name = 'color_dmd'



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAnimation(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/animation'

    def get_config_file(self):
        return 'test_animation.yaml'

    def test_animation_config_processing(self):
        # The animation sections are more complex than regular sections,
        # so they do some more complex pre-processing before they go to the
        # config file validator.

        # slide def, list of animations
        s1w0 = self.mc.slides['slide1']['widgets'][0]['animations']

        self.assertIs(type(s1w0['show_slide']), list)
        self.assertEqual(len(s1w0['show_slide']), 2)
        self.assertIs(type(s1w0['show_slide'][0]), dict)
        self.assertIs(type(s1w0['show_slide'][1]), dict)
        self.assertEqual(s1w0['show_slide'][0]['value'], ['101'])
        self.assertEqual(s1w0['show_slide'][1]['value'], ['100'])

        # slide def, single dict animation
        s2w0 = self.mc.slides['slide2']['widgets'][0]['animations']

        self.assertIs(type(s2w0['entrance2']), list)
        self.assertEqual(len(s2w0['entrance2']), 1)
        self.assertIs(type(s2w0['entrance2'][0]), dict)
        self.assertEqual(s2w0['entrance2'][0]['value'], ['0' ,'0'])
        self.assertEqual(s2w0['entrance2'][0]['property'], ['x', 'y'])

        # slide def, 1 event, list of 2 named animations
        s3w0 = self.mc.slides['slide3']['widgets'][0]['animations']
        self.assertIs(type(s3w0['entrance3']), list)
        self.assertEqual(len(s3w0['entrance3']), 2)
        self.assertIs(type(s3w0['entrance3'][0]), dict)
        self.assertIs(type(s3w0['entrance3'][1]), dict)
        self.assertEqual(s3w0['entrance3'][0]['named_animation'], 'fade_in')
        self.assertEqual(s3w0['entrance3'][1]['named_animation'], 'multi')

        # slide def, 2 events, list of named animations
        s4w0 = self.mc.slides['slide4']['widgets'][0]['animations']
        self.assertIs(type(s4w0['entrance4']), list)
        self.assertEqual(len(s4w0['entrance4']), 2)
        self.assertIs(type(s4w0['entrance4'][0]), dict)
        self.assertIs(type(s4w0['entrance4'][1]), dict)
        self.assertEqual(s4w0['entrance4'][0]['named_animation'], 'fade_in')
        self.assertEqual(s4w0['entrance4'][1]['named_animation'], 'multi')

        self.assertIs(type(s4w0['some_event4']), list)
        self.assertEqual(len(s4w0['some_event4']), 1)
        self.assertIs(type(s4w0['some_event4'][0]), dict)
        self.assertEqual(s4w0['some_event4'][0]['named_animation'], 'multi')

        # slide def, 2 events, 1 named animation, 1 dict
        s5w0 = self.mc.slides['slide5']['widgets'][0]['animations']
        self.assertIs(type(s5w0['entrance5']), list)
        self.assertEqual(len(s5w0['entrance5']), 2)
        self.assertIs(type(s5w0['entrance5'][0]), dict)
        self.assertIs(type(s5w0['entrance5'][1]), dict)
        self.assertEqual(s5w0['entrance5'][0]['named_animation'], 'fade_in')
        self.assertEqual(s5w0['entrance5'][1]['named_animation'], 'multi')

        self.assertIs(type(s5w0['event5']), list)
        self.assertEqual(len(s5w0['event5']), 1)
        self.assertIs(type(s5w0['event5'][0]), dict)
        self.assertEqual(s5w0['event5'][0]['value'], ['98'])

        # slide with 1 widget with no animations
        self.assertIn('animations', self.mc.slides['slide6']['widgets'][0])
        self.assertIsNone(self.mc.slides['slide6']['widgets'][0]['animations'])

        # Move on to test the named animations section

        self.assertEqual(len(self.mc.animations), 6)

        # single animation, dict
        self.assertIs(type(self.mc.animations['fade_in']), list)
        self.assertEqual(len(self.mc.animations['fade_in']), 1)
        self.assertIs(type(self.mc.animations['fade_in'][0]), dict)
        self.assertEqual(self.mc.animations['fade_in'][0]['property'],
                         ['opacity'])
        self.assertEqual(self.mc.animations['fade_in'][0]['easing'],
                         'linear')

        # two animations, list, with values as percent strings
        self.assertIs(type(self.mc.animations['multi']), list)
        self.assertEqual(len(self.mc.animations['multi']), 2)
        self.assertIs(type(self.mc.animations['multi'][0]), dict)
        self.assertEqual(self.mc.animations['multi'][0]['property'],
                         ['y'])
        self.assertEqual(self.mc.animations['multi'][0]['easing'],
                         'linear')
        self.assertFalse(self.mc.animations['multi'][0]['relative'])
        self.assertIs(type(self.mc.animations['multi'][1]), dict)
        self.assertEqual(self.mc.animations['multi'][1]['property'],
                         ['x'])
        self.assertEqual(self.mc.animations['multi'][1]['easing'],
                         'linear')
        self.assertFalse(self.mc.animations['multi'][1]['relative'])

        self.assertIs(type(self.mc.animations['advance_x_50']), list)
        self.assertEqual(len(self.mc.animations['advance_x_50']), 1)
        self.assertIs(type(self.mc.animations['advance_x_50'][0]), dict)
        self.assertEqual(self.mc.animations['advance_x_50'][0]['property'],
                         ['x'])
        self.assertEqual(self.mc.animations['advance_x_50'][0]['easing'],
                         'linear')
        self.assertTrue(self.mc.animations['advance_x_50'][0]['relative'])

        self.assertIs(type(self.mc.animations['advance_y_50']), list)
        self.assertEqual(len(self.mc.animations['advance_y_50']), 1)
        self.assertIs(type(self.mc.animations['advance_y_50'][0]), dict)
        self.assertEqual(self.mc.animations['advance_y_50'][0]['property'],
                         ['y'])
        self.assertEqual(self.mc.animations['advance_y_50'][0]['easing'],
                         'linear')
        self.assertTrue(self.mc.animations['advance_y_50'][0]['relative'])

        self.assertIs(type(self.mc.animations['advance_xy_50']), list)
        self.assertEqual(len(self.mc.animations['advance_xy_50']), 1)
        self.assertIs(type(self.mc.animations['advance_xy_50'][0]), dict)
        self.assertIs(type(self.mc.animations['advance_xy_50'][0]['property']), list)
        self.assertIn('x', self.mc.animations['advance_xy_50'][0]['property'])
        self.assertIn('y', self.mc.animations['advance_xy_50'][0]['property'])
        self.assertEqual(self.mc.animations['advance_xy_50'][0]['easing'],
                         'linear')
        self.assertTrue(self.mc.animations['advance_xy_50'][0]['relative'])

    def test_reset_animations_pre_show_slide(self):
        self.mc.events.post('show_slide13')
        self.advance_time(.1)

        widget = self.mc.active_slides['slide13'].children[0].children[0]

        self.assertAlmostEqual(-140, widget.anchor_offset_pos[0], delta=20)
        self.assertEqual(100, widget.x)
        self.advance_time(.5)
        self.assertAlmostEqual(150, widget.x, delta=20)
        self.advance_time(.5)
        self.assertEqual(200.0, widget.x)

        self.mc.events.post('show_base_slide')
        self.advance_time()

        # animations should start from orig position
        self.mc.events.post('show_slide13')
        self.advance_time(.1)

        # refetch widget because this is another slide instance
        widget = self.mc.active_slides['slide13'].children[0].children[0]

        self.assertAlmostEqual(-140, widget.anchor_offset_pos[0], delta=20)
        self.assertEqual(100, widget.x)
        self.advance_time(.5)
        self.assertAlmostEqual(150, widget.x, delta=20)
        self.advance_time(.5)
        self.assertEqual(200.0, widget.x)

    def test_reset_animations_slide_play(self):
        self.mc.events.post('show_slide14')
        self.advance_time(.1)

        widget = self.mc.active_slides['slide14'].children[0].children[0]

        self.assertAlmostEqual(-120, widget.anchor_offset_pos[0], delta=20)
        self.assertEqual(100, widget.x)
        self.advance_time(.5)
        self.assertAlmostEqual(150, widget.x, delta=20)
        self.advance_time(.5)
        self.assertEqual(200.0, widget.x)

        self.mc.events.post('show_base_slide')
        self.advance_time()

        # animations should start from orig position
        self.mc.events.post('show_slide14')
        self.advance_time(.1)

        # refetch widget because this is another slide instance
        widget = self.mc.active_slides['slide14'].children[0].children[0]

        self.assertAlmostEqual(-120, widget.anchor_offset_pos[0], delta=20)
        self.assertEqual(100, widget.x)
        self.advance_time(.5)
        self.assertAlmostEqual(150, widget.x, delta=20)
        self.advance_time(.5)
        self.assertEqual(200.0, widget.x)

    def test_reset_animations_standard_event(self):
        self.mc.events.post('show_slide15')
        self.advance_time(.1)

        widget = self.mc.targets['default'].current_slide.widgets[0].widget

        self.assertAlmostEqual(-130, widget.anchor_offset_pos[0], delta=20)
        self.assertEqual(100, widget.x)
        self.advance_time(.5)
        self.assertAlmostEqual(150, widget.x, delta=20)
        self.advance_time(.5)
        self.assertEqual(200.0, widget.x)

        self.mc.events.post('event1')
        self.advance_time()
        self.assertEqual(100, widget.x)

    def test_animation_show_slide(self):
        self.mc.events.post('show_slide7')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name, 'slide7')

        # grab this widget
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'TEST ANIMATION ON show_slide')

        # make sure it's animating as we expect
        self.assertTrue(widget.animation)
        self.assertEqual(widget.animation.duration, 0.5)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties, {'x': 500})

        # wait for it to finish and make sure it stopped
        self.advance_time(.5)
        self.assertIsNone(widget.animation.have_properties_to_animate(widget))

    def test_animation_show_slide_with_transition(self):
        self.mc.events.post('show_base_slide')
        self.advance_time()

        self.mc.events.post('show_slide10')
        self.advance_time()

        # transition is 1 sec, so animation should not start until that's done
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'ANIMATION show_slide')
        self.assertFalse(widget.animation)

        self.advance_time(1)
        # slide should be fully transitioned in, so animation should be started

        self.assertTrue(widget.animation)
        self.assertEqual(widget.animation.duration, 0.5)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties, {'x': 500})

    def test_animation_pre_show_slide(self):
        self.mc.events.post('show_base_slide')
        self.advance_time()

        self.mc.events.post('show_slide9')
        self.advance_time()

        # transition is 1 sec, but animation should run at the same time
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'ANIMATION pre_show_slide')

        self.assertTrue(widget.animation)
        self.assertEqual(widget.animation.duration, 0.5)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties, {'x': 500})

    def test_animation_pre_slide_leave(self):
        self.mc.events.post('show_slide11')
        self.advance_time()

        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'ANIMATION pre_slide_leave')

        self.mc.events.post('show_base_slide_with_transition')
        self.advance_time()

        # transition is 1 sec, but leave animation should run at the same time
        self.assertTrue(widget.animation)
        self.assertEqual(widget.animation.duration, 0.5)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties, {'x': -400})

        # todo this test passes and the events fire, but visually it doesn't
        # seem to work. Bug in kivy? Or maybe it's by design?

    def test_animation_slide_leave(self):
        self.mc.events.post('show_slide12')
        self.advance_time()

        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'ANIMATION slide_leave')
        self.assertFalse(widget.animation)

        self.mc.events.post('show_base_slide_with_transition')
        self.advance_time()

        self.advance_time(1)
        # slide should be fully transitioned out, so animation should be
        # started

        self.assertTrue(widget.animation)
        self.assertEqual(widget.animation.duration, 0.5)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties, {'x': 0})

    def test_animation_from_event(self):
        # This will also test multiple animated properties on the same entry

        self.mc.events.post('show_slide2')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        # grab this widget
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'ANIMATION TEST')

        # make sure it's not animating
        self.assertIsNone(widget.animation)

        # post the event to animate it
        self.mc.events.post('entrance2')
        self.advance_time()

        # make sure it's doing what we expect
        self.assertEqual(widget.animation.duration, 1)
        self.assertTrue(widget.animation.have_properties_to_animate(widget))
        self.assertEqual(widget.animation.animated_properties,
                         {'x': 0, 'y': 0})
        self.advance_time(.5)
        self.assertAlmostEqual(190, widget.x, delta=30)
        self.assertAlmostEqual(145, widget.y, delta=30)

    def test_named_animation(self):
        self.mc.events.post('show_slide3')

        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name, 'slide3')

        # grab this widget
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'text3')

        # make sure it's not animating
        self.assertIsNone(widget.animation)

        # make sure initial values are set
        self.assertEqual(widget.opacity, 0)
        self.assertEqual(widget.x, 200)
        self.assertEqual(widget.y, 150)

        # post the event to animate it
        self.mc.events.post('entrance3')
        self.advance_time(1.1)

        # post-animation opacity should be 1
        self.assertEqual(widget.opacity, 1)

        # advance past animation step 2
        self.advance_time(2.1)

        # check properties
        self.assertEqual(widget.x, 0)
        self.assertEqual(widget.y, 0)

        # add widget2
        self.mc.events.post('show_widget2')
        self.advance_time()

        # grab this widget
        widget = self.mc.targets['default'].current_slide.children[1].widget
        self.assertTrue(widget.text == 'widget2')

        # make sure it's not animating
        self.assertIsNone(widget.animation)

        # make sure initial values are set
        self.assertEqual(widget.opacity, 0)
        self.assertEqual(widget.x, 200)
        self.assertEqual(widget.y, 150)

        # post the event to animate it
        self.mc.events.post('animate_widget2')
        self.advance_time(1.1)

        # post-animation opacity should be 1
        self.assertEqual(widget.opacity, 1)

        # advance past animation step 2
        self.advance_time(2.1)

        # check properties
        self.assertEqual(widget.x, 0)
        self.assertEqual(widget.y, 0)

        self.mc.events.post('pulse_widget2')
        self.advance_time(1)

    def test_add_to_slide_from_offscreen(self):
        self.mc.events.post('show_slide8')
        self.advance_time()

        self.mc.events.post('show_widget1')
        self.advance_time()

        widget = self.mc.targets['default'].current_slide.children[1].widget
        self.assertEqual('WIDGET 1', widget.text)
        self.assertEqual(widget.pos[0], -100)

        self.mc.events.post('move_on_slide')
        self.advance_time(.6)
        self.assertEqual(widget.pos[0], 100)

        self.mc.events.post('move_off_slide')
        self.advance_time(.6)
        self.assertEqual(widget.pos[0], -100)

    def test_relative_animation(self):
        self.mc.events.post('show_slide3')

        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name, 'slide3')

        # grab this widget
        widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertTrue(widget.text == 'text3')

        # make sure it's not animating
        self.assertIsNone(widget.animation)

        # make sure initial values are set
        self.assertEqual(widget.opacity, 0)
        self.assertEqual(widget.x, 200)
        self.assertEqual(widget.y, 150)

        # post the event to fade it in
        self.mc.events.post('fade_in')
        self.advance_time(1.1)

        # post-animation opacity should be 1
        self.assertEqual(widget.opacity, 1)

        # post the event to advance it +50 in the x direction
        self.mc.events.post('advance_x')
        self.advance_time(1.1)

        # check properties
        self.assertEqual(widget.x, 250)
        self.assertEqual(widget.y, 150)

        # post the event to advance it +50 in the y direction
        self.mc.events.post('advance_y')
        self.advance_time(1.1)

        # check properties
        self.assertEqual(widget.x, 250)
        self.assertEqual(widget.y, 200)

        # post the event to advance it +50 in both the x and y directions
        self.mc.events.post('advance_xy')
        self.advance_time(1.1)

        # check properties
        self.assertEqual(widget.x, 300)
        self.assertEqual(widget.y, 250)



"""BCP Settings Controller for MC."""
from mpf.core.settings_controller import SettingsController, SettingEntry


class McSettingsController(SettingsController):

    """MC Settings Controller.

    Gets settings via BCP. Can also set settings using bcp.
    """

    def _add_entries_from_config(self):
        """Do not load entries from config."""

    def add_setting(self, setting):
        """Add a setting."""
        setting = SettingEntry(*setting)
        self._settings[setting.name] = setting



from kivy.uix.effectwidget import EffectBase
from kivy.properties import NumericProperty


class GammaEffect(EffectBase):
    """GLSL effect to apply a gamma setting to a texture"""

    gamma = NumericProperty(1.0)
    '''
    Sets the gamma factor.

    gamma is a :class:`~kivy.properties.NumericProperty` and
    defaults to 1.0 (which has no effect).
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_gamma(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = gamma_glsl.format(float(self.gamma))


gamma_glsl = '''
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    vec4 outColor = vec4(pow(color.x, {0}), pow(color.y, {0}), pow(color.z, {0}), 1.0);
    return outColor;
}}
'''

effect_cls = GammaEffect
name = 'gamma'



"""Widget showing a video."""
from typing import Optional

from kivy.uix.video import Video
from kivy.core.video import Video as CoreVideo
from kivy.graphics import Rectangle, Color, Rotate, Scale
from kivy.properties import NumericProperty

from mpfmc.uix.widget import Widget, magic_events

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc
    from kivy.graphics.texture import Texture   # noqa


class VideoWidget(Widget, Video):

    """Widget showing a video."""

    widget_type_name = 'Video'
    merge_settings = ('height', 'width')
    animation_properties = ('x', 'y')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs

        super().__init__(mc=mc, config=config, key=key)

        try:
            self.video = self.mc.videos[self.config['video']]
        except Exception:
            raise ValueError("Cannot add Video widget. Video '{}' is not a "
                             "valid video name.".format(self.config['video']))

        self._control_events = list()

        self._registered_magic_events = dict()
        for event in magic_events:
            self._registered_magic_events[event] = list()

        self.merge_asset_config(self.video)

        if self.config['control_events']:
            self._setup_control_events(self.config['control_events'])

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(pos=self._draw_widget,
                  size=self._draw_widget,
                  color=self._draw_widget,
                  texture=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget)

        if not self.video.video:
            self.video.load(callback=self._do_video_load)
        else:
            self._do_video_load()

    def __repr__(self) -> str:  # pragma: no cover
        try:
            return '<Video name={}, size={}, pos={}>'.format(self.video.name,
                                                             self.size,
                                                             self.pos)
        except AttributeError:
            return '<Video (loading...), size={}, pos={}>'.format(self.size,
                                                                  self.pos)

    def _draw_widget(self, *args):
        """Draws the image (draws a rectangle using the image texture)"""
        del args

        anchor = (self.x - self.anchor_offset_pos[0], self.y - self.anchor_offset_pos[1])
        self.canvas.clear()

        if self.state in ('play', 'pause'):
            with self.canvas:
                Color(*self.color)
                Rotate(angle=self.rotation, origin=anchor)
                Scale(self.scale).origin = anchor
                Rectangle(pos=self.pos, size=self.size, texture=self.texture)

    def _setup_control_events(self, event_list: list) -> None:
        for entry in event_list:

            kwargs = dict()

            if entry['action'] == 'play':
                handler = self.play

            elif entry['action'] == 'pause':
                handler = self.pause

            elif entry['action'] == 'stop':
                handler = self.stop

            elif entry['action'] == 'seek':
                handler = self.seek
                kwargs = {'percent': entry['value']}

            elif entry['action'] == 'volume':
                handler = self.set_volume
                kwargs = {'volume': entry['value']}

            elif entry['action'] == 'position':
                handler = self.set_playback_position
                kwargs = {'position': entry['value']}

            else:
                raise AssertionError("Invalid control_event action {} in "
                                     "video".format(entry['action']), self)

            if entry['event'] in magic_events:
                self._registered_magic_events[entry['event']].append(
                    (handler, kwargs))

            else:
                self._control_events.append(
                    self.mc.events.add_handler(entry['event'], handler,
                                               **kwargs))

    def on_add_to_slide(self, dt) -> None:
        super().on_add_to_slide(dt)
        for handler, kwargs in self._registered_magic_events['add_to_slide']:
            handler(**kwargs)

    def on_remove_from_slide(self) -> None:
        super().on_remove_from_slide()
        for handler, kwargs in self._registered_magic_events[
                'remove_from_slide']:
            handler(**kwargs)

    def on_pre_show_slide(self) -> None:
        super().on_pre_show_slide()
        for handler, kwargs in self._registered_magic_events['pre_show_slide']:
            handler(**kwargs)

    def on_show_slide(self) -> None:
        super().on_show_slide()
        for handler, kwargs in self._registered_magic_events['show_slide']:
            handler(**kwargs)

    def on_pre_slide_leave(self) -> None:
        super().on_pre_slide_leave()
        for handler, kwargs in self._registered_magic_events[
                'pre_slide_leave']:
            handler(**kwargs)

    def on_slide_leave(self) -> None:
        super().on_slide_leave()
        for handler, kwargs in self._registered_magic_events['slide_leave']:
            handler(**kwargs)

    def on_slide_play(self) -> None:
        super().on_slide_play()
        for handler, kwargs in self._registered_magic_events['slide_play']:
            handler(**kwargs)

    def play(self, **kwargs) -> None:
        del kwargs
        if not self.video.loaded:
            return
        self.state = 'play'

    def pause(self, **kwargs) -> None:
        del kwargs
        if not self.video.loaded:
            return
        self.state = 'pause'

    def stop(self, **kwargs) -> None:
        del kwargs
        if not self.video.loaded:
            return
        self.state = 'stop'

    def seek(self, percent, **kwargs) -> None:  # noqa
        if not self.video.loaded:
            return
        super().seek(percent, **kwargs)

    def set_volume(self, volume, **kwargs) -> None:
        del kwargs
        if not self.video.loaded:
            return
        self.volume = volume

    def set_playback_position(self, position: int, **kwargs) -> None:
        del kwargs
        if not self.video.loaded:
            return
        super().seek(position / self.duration)

    def _do_video_load(self, *largs) -> None:
        # Overrides a method in the base Kivy Video widget. It's basically
        # copied and pasted from there, with the change that this method pulls
        # the video object from the MPF asset versus loading from a file and
        # it has the extra bit at the end to set the size and position

        del largs

        if CoreVideo is None:
            raise TypeError("Could not find a video provider to play back "
                            "the video '{}'".format(self.video.file))

        self.volume = self.config['volume']
        if self.config['auto_play']:
            self.state = 'play'
        else:
            self.state = 'stop'

        # pylint: disable-msg=access-member-before-definition
        # Disable the error about accessing an attribute before its defined.
        # self._video is defined in the base class, but not in a direct way, so
        # pylint doesn't see it.
        if self._video:
            self._video.stop()

        self._video = self.video.video
        self._video.volume = self.volume
        self._video.bind(on_load=self._on_load,
                         on_frame=self._on_video_frame,
                         on_eos=self._on_eos)
        # This is also flagged as an error by pylint, but it's okay because
        # self.state is defined in the base class.

        if self.state == 'play':
            self._video.play()
        self.duration = 1.

        self.video.set_end_behavior(self.config['end_behavior'])

    def on_texture(self, instance, value: "Texture") -> None:
        # Overrides the base method to put the size into self.size instead of
        # self.texture_size
        del instance

        if value is not None:
            if self.config['width'] and self.config['height']:
                self.size = (self.config['width'], self.config['height'])
            else:
                self.size = list(value.size)

    def prepare_for_removal(self) -> None:
        super().prepare_for_removal()
        self.mc.events.remove_handlers_by_keys(self._control_events)
        self._control_events = list()
        self.stop()

    #
    # Properties
    #

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [VideoWidget]



import zipfile
from io import BytesIO

from kivy import Logger

from kivy.cache import Cache

from kivy.core.image import Image, ImageLoaderBase, ImageLoader, Texture
from mpf.core.assets import AssetPool

from mpfmc.assets.mc_asset import McAsset

# This module has extra comments since it's what we tell people to use as an
# example of an Asset implementation.


class ImagePool(AssetPool):

    """A pool of images."""

    __slots__ = []

    def __repr__(self):
        # String that's returned if someone prints this object
        return '<ImagePool: {}>'.format(self.name)

    @property
    def image(self):
        return self.asset


class LazyZipImageLoaderTexture:

    """Lazy textures for images inside a zip."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, zip_file, filename, mipmap, keep_data, no_cache):
        self._zip_file = zip_file
        self._mipmap = mipmap
        self._keep_data = keep_data
        self._no_cache = no_cache
        self._filename = filename
        znamelist = self._zip_file.namelist()
        znamelist.sort()
        self._index_list = []
        self.width = None
        self.height = None

        for zfilename in znamelist:
            self._index_list.append(zfilename)

        self._loaded_textures = [None] * len(self._index_list)

    def __len__(self):
        return len(self._index_list)

    def __getitem__(self, item):
        if not self._loaded_textures[item]:
            # first, check if a texture with the same name already exist in the
            # cache
            # pylint: disable-msg=redefined-builtin
            chr = type(self._filename)
            uid = chr(u'%s|%d|%d') % (self._filename, self._mipmap, item)
            texture = Cache.get('kv.texture', uid)

            # if not create it and append to the cache
            if texture is None:
                zfilename = self._index_list[item]
                # read file and store it in mem with fileIO struct around it
                tmpfile = BytesIO(self._zip_file.read(zfilename))
                ext = zfilename.split('.')[-1].lower()
                image = None
                for loader in ImageLoader.loaders:
                    if (ext not in loader.extensions() or
                            not loader.can_load_memory()):
                        continue
                    Logger.debug('Image%s: Load <%s> from <%s>',
                                 loader.__name__[11:], zfilename,
                                 self._filename)
                    try:
                        image = loader(zfilename, ext=ext, rawdata=tmpfile,
                                       inline=True)
                    except:     # pylint: disable-msg=bare-except   # noqa
                        # Loader failed, continue trying.
                        continue
                    break
                if image is None:
                    raise AssertionError("Could not load image {} (index {}) "
                                         "from zip {}".format(zfilename, item,
                                                              self._filename))

                self.width = image.width
                self.height = image.height

                imagedata = image._data[0]  # pylint: disable-msg=protected-access

                source = '{}{}|'.format(
                    'zip|' if self._filename.endswith('.zip') else '',
                    self._no_cache)
                imagedata.source = chr(source) + uid
                texture = Texture.create_from_data(
                    imagedata, mipmap=self._mipmap)
                if not self._no_cache:
                    Cache.append('kv.texture', uid, texture)
                if imagedata.flip_vertical:
                    texture.flip_vertical()

            self._loaded_textures[item] = texture

        return self._loaded_textures[item]


class LazyZipImageLoader(ImageLoaderBase):

    """Lazy image loader for image inside a zip."""

    @staticmethod
    def save(*largs, **kwargs):
        raise AssertionError("Not supported")

    def __init__(self, filename, zip_file, **kwargs):
        super().__init__(filename, **kwargs)
        self._zipfile = zip_file
        self._data = dict()     # to prevent breakage in loader::_load_urllib
        self._textures = None

    def load(self, filename):
        """Return the zip object."""
        return filename

    def populate(self):
        """Populate textures with lazy loader."""
        if not self._textures:
            self._textures = LazyZipImageLoaderTexture(self._zipfile,
                                                       self.filename,
                                                       self._mipmap,
                                                       self.keep_data,
                                                       self._nocache)

    @property
    def width(self):
        '''Image width
        '''
        if not self._textures:
            self.populate()
        return self._textures.width

    @property
    def height(self):
        '''Image height
        '''
        if not self._textures:
            self.populate()
        return self._textures.height

    @property
    def size(self):
        '''Image size (width, height)
        '''
        return (self.width, self.height)


class KivyImageLoaderPatch:

    """Patch Kivy zip loader."""

    @staticmethod
    def lazy_zip_loader(filename):
        '''Read images from an zip file lazily.

        .. versionadded:: 1.12.0

        Returns an LazyZipImageLoader which loads images from a zip on demand.
        '''
        # read zip in memory for faster access
        with open(filename, 'rb') as handle:
            _file = BytesIO(handle.read())
        # read all images inside the zip
        zip_file = zipfile.ZipFile(_file)

        return LazyZipImageLoader(filename, zip_file=zip_file, inline=True)


class ImageAsset(McAsset):

    attribute = 'images'  # attribute in MC, e.g. self.mc.images
    path_string = 'images'  # entry from mpf-mc:paths: for asset folder name
    config_section = 'images'  # section in the config files for this asset
    extensions = ('png', 'jpg', 'jpeg', 'gif', 'zip', 'bmp')  # obvious. No dots.
    class_priority = 100  # Order asset classes are loaded. Higher is first.
    pool_config_section = 'image_pools'  # Will setup groups if present
    asset_group_class = ImagePool  # Class or None to not use pools

    __slots__ = ["references", "_image"]

    def __init__(self, mc, name, file, config):
        super().__init__(mc, name, file, config)  # be sure to call super

        # Do whatever else you want here. You can remove the entire method if
        # you don't need to do anything.

        self._image = None  # holds the actual image in memory
        self.references = 0

    @property
    def image(self):
        # Since self._image will change depending on whether the image is
        # loaded or not, set a property so external methods can just use
        # ImageAsset.image

        return self._image

    def do_load(self):
        """Load the image."""
        # This is the method that's actually called to load the asset from
        # disk. It's called by the loader thread so it's ok to block. However
        # since it's a separate thread, don't update any other attributes.

        # When you're done loading and return, the asset will be processed and
        # the various load status attributes will be updated automatically,
        # and anything that was waiting for it to load will be called. So
        # all you have to do here is load and return.

        if self.machine.machine_config['mpf-mc']['zip_lazy_loading']:
            # lazy loading for zip file image sequences
            ImageLoader.zip_loader = KivyImageLoaderPatch.lazy_zip_loader

        self._image = Image(self.config['file'],
                            keep_data=False,
                            scale=1.0,
                            mipmap=False,
                            anim_delay=-1,
                            nocache=True)

        self._image.anim_reset(False)

        # load first texture to speed up first display
        self._callbacks.add(lambda x: self._image.texture)

    def _do_unload(self):
        # This is the method that's called to unload the asset. It's called by
        # the main thread so you don't have to worry about thread
        # complexities, but since it's in the main thread, you need to
        # return quickly.

        self._image = None



# ---------------------------------------------------------------------------
#    Definitions from SDL2 library
# ---------------------------------------------------------------------------

cdef extern from "Python.h":
    void PyEval_InitThreads()

cdef extern from "stdio.h" nogil:
    cdef struct __sFILE:
        pass
    ctypedef __sFILE FILE

# ---------------------------------------------------------------------------
#    SDL2
# ---------------------------------------------------------------------------
cdef extern from "SDL.h" nogil:

    ctypedef unsigned char Uint8
    ctypedef signed char Sint8
    ctypedef unsigned long Uint32
    ctypedef signed long Sint32
    ctypedef unsigned long long Uint64
    ctypedef signed long long Sint64
    ctypedef signed short Sint16
    ctypedef unsigned short Uint16
    ctypedef Uint16 SDL_AudioFormat

    int SDL_AUDIO_ALLOW_FREQUENCY_CHANGE
    int SDL_AUDIO_ALLOW_FORMAT_CHANGE
    int SDL_AUDIO_ALLOW_CHANNELS_CHANGE
    int SDL_AUDIO_ALLOW_ANY_CHANGE

    enum Enum_temp_random_970738:
        SDL_FALSE
        SDL_TRUE
    ctypedef Enum_temp_random_970738 SDL_bool

    int SDL_INIT_AUDIO
    int AUDIO_S16SYS

    int SDL_MIX_MAXVOLUME

    bint SDL_AUDIO_ISLITTLEENDIAN(SDL_AudioFormat format)
    int SDL_AUDIO_BITSIZE(SDL_AudioFormat)

    struct SDL_AudioSpec:
        int freq
        SDL_AudioFormat format
        Uint8 channels
        Uint8 silence
        Uint16 samples
        Uint16 padding
        Uint32 size
        void (*callback)(void *userdata, Uint8 *stream, int len)
        void *userdata

    struct SDL_mutex:
        pass

    struct SDL_Thread:
        pass

    ctypedef Uint32 SDL_AudioDeviceID
    SDL_AudioDeviceID SDL_OpenAudioDevice(char *device, int iscapture, SDL_AudioSpec *desired, SDL_AudioSpec *obtained, int allowed_changes)
    void SDL_LockAudioDevice(SDL_AudioDeviceID dev)
    void SDL_UnlockAudioDevice(SDL_AudioDeviceID dev)
    cdef enum Enum_temp_random_404053:
        SDL_AUDIO_STOPPED
        SDL_AUDIO_PLAYING
        SDL_AUDIO_PAUSED
    ctypedef Enum_temp_random_404053 SDL_AudioStatus
    SDL_AudioStatus SDL_GetAudioStatus()
    SDL_AudioStatus SDL_GetAudioDeviceStatus(SDL_AudioDeviceID dev)
    void SDL_PauseAudioDevice(SDL_AudioDeviceID dev, int pause_on)
    void SDL_CloseAudioDevice(SDL_AudioDeviceID dev)

    SDL_mutex *SDL_CreateMutex()
    void SDL_DestroyMutex(SDL_mutex *)
    int SDL_LockMutex(SDL_mutex *)
    int SDL_UnlockMutex(SDL_mutex *)

    struct SDL_cond:
        pass

    SDL_cond *SDL_CreateCond()
    void SDL_DestroyCond(SDL_cond *)
    int SDL_CondSignal(SDL_cond *)
    int SDL_CondWait(SDL_cond *, SDL_mutex *)

    struct SDL_Thread:
        pass

    ctypedef int (*SDLCALL)(void *)
    SDL_Thread *SDL_CreateThread(SDLCALL, void *data)
    void SDL_WaitThread(SDL_Thread *thread, int *status)
    Uint32 SDL_ThreadID()

    void SDL_SetError(char *fmt)
    char *SDL_GetError()
    void SDL_ClearError()

    struct SDL_UserEvent:
        Uint8 type
        int code
        void *data1
        void *data2

    union SDL_Event:
        Uint8 type

    int SDL_PushEvent(SDL_Event *event)
    void SDL_Delay(int)
    int SDL_Init(int)
    int SDL_InitSubSystem(Uint32 flags)
    void SDL_Quit()
    void SDL_LockAudio()
    void SDL_UnlockAudio()
    void SDL_MixAudio(Uint8 *dst, const Uint8 *src, Uint32 len, int volume)
    void SDL_MixAudioFormat(Uint8 *dst, const Uint8 *src, SDL_AudioFormat format,
                                                Uint32 len, int volume)
    SDL_AudioSpec* SDL_LoadWAV(const char* file, SDL_AudioSpec* spec, Uint8** audio_buf, Uint32* audio_len)
    void SDL_FreeWAV(Uint8* audio_buf)
    void SDL_free(void *ptr)

    Uint32 SDL_GetTicks()

    struct SDL_version:
        Uint8 major
        Uint8 minor
        Uint8 patch

    void SDL_GetVersion(SDL_version *ver)

    struct SDL_AudioCVT:
        pass
    ctypedef void (*SDL_AudioFilter)(SDL_AudioCVT* cvt, SDL_AudioFormat format)
    struct SDL_AudioCVT:
        int needed
        SDL_AudioFormat src_format
        SDL_AudioFormat dst_format
        double rate_incr
        Uint8 *buf
        int len
        int len_cvt
        int len_mult
        double len_ratio
        SDL_AudioFilter filters[10]
        int filter_index

    int SDL_BuildAudioCVT(SDL_AudioCVT *cvt, SDL_AudioFormat src_fmt, Uint8 src_channels, int src_rate,
                          SDL_AudioFormat dst_fmt, Uint8 dst_channels, int dst_rate)
    int SDL_ConvertAudio(SDL_AudioCVT *cvt)

    int RW_SEEK_SET
    int RW_SEEK_CUR
    int RW_SEEK_END

    struct Struct_temp_random_697984:
        SDL_bool autoclose
        FILE *fp
    struct Struct_temp_random_614396:
        Uint8 *base
        Uint8 *here
        Uint8 *stop
    struct Struct_temp_random_176105:
        void *data1
    union Union_temp_random_505832:
        Struct_temp_random_697984 stdio
        Struct_temp_random_614396 mem
        Struct_temp_random_176105 unknown

    struct SDL_RWops:
        long (*seek)(SDL_RWops* context, long offset, int whence) nogil
        unsigned long (*read)(SDL_RWops* context, void* ptr, size_t size, size_t maxnum) nogil
        unsigned long (*write)(SDL_RWops* context, void* ptr, size_t size, size_t num) nogil
        int (*close)(SDL_RWops* context) nogil
        Uint32 type
        Union_temp_random_505832 hidden

    long SDL_RWseek(SDL_RWops *context, long offset, int whence)
    long SDL_RWtell(SDL_RWops *context)
    unsigned long SDL_RWread(SDL_RWops *context, void* ptr, size_t size, size_t maxnum)
    int SDL_RWclose(SDL_RWops *context)

    SDL_RWops *SDL_RWFromFile(const char *file, const char *mode)

# ---------------------------------------------------------------------------
#    SDL2 helper functions defined in sdl2_helper.h
# ---------------------------------------------------------------------------
cdef extern from 'sdl2_helper.h':
    int convert_audio_to_desired_format(SDL_AudioSpec input_spec, SDL_AudioSpec desired_spec,
                                        Uint8* input_buffer, Uint32 input_size, Uint8** output_buffer, Uint32 *output_size)


# ---------------------------------------------------------------------------
#    SDL_Mixer
# ---------------------------------------------------------------------------
cdef extern from "SDL_mixer.h" nogil:

    cdef enum MIX_InitFlags:
        MIX_INIT_FLAC,
        MIX_INIT_MP3,
        MIX_INIT_OGG

    struct Mix_Chunk:
        int allocated
        Uint8 *abuf
        Uint32 alen
        Uint8 volume

    int MIX_MAX_VOLUME

    void Mix_Quit()
    int Mix_OpenAudio(int frequency, Uint16 format, int channels, int chunksize)
    void Mix_CloseAudio()
    int Mix_QuerySpec(int *frequency, Uint16 *format, int *channels)
    const SDL_version *Mix_Linked_Version()
    int Mix_AllocateChannels(int numchans)
    void Mix_HookMusic(void (*mix_func)(void *udata, Uint8 *stream, int len), void *arg)
    void *Mix_GetMusicHookData()
    Mix_Chunk *Mix_LoadWAV(char *file)


# ---------------------------------------------------------------------------
#    Helper structures
# ---------------------------------------------------------------------------

# The following declarations are not from SDL/SDL_Mixer, but are application-
# specific data structures used in the MPF media controller audio library:

cdef struct Sample16Bytes:
    # Structure that represents two bytes of a 16-bit sample.  This is used in
    # the union below (nested structs are not permitted in Cython so it is
    # declared separately here).
    Uint8 byte0
    Uint8 byte1

cdef union Sample16:
    # Union structure that represents a single 16-bit sample value.  A union is
    # utilized to make it easy to access the individual bytes in the sample.
    # This is needed since all samples are streamed as individual 8-bit (byte)
    # values.
    Sint16 value
    Sample16Bytes bytes


# ---------------------------------------------------------------------------
#    Audio Callback Data type
# ---------------------------------------------------------------------------

ctypedef struct AudioCallbackData:
    # A pointer to this struct is passed to the main audio callback function and
    # is the only way data is made available to the main audio thread.  Must not
    # contain any Python objects.
    SDL_AudioFormat format
    int sample_rate
    int channels
    Uint16 buffer_samples
    Uint32 buffer_size
    Uint16 bytes_per_control_point
    Uint8 bytes_per_sample
    double seconds_to_bytes_factor
    Uint8 quick_fade_steps
    Uint8 master_volume
    Uint8 silence
    Uint8 track_count
    void **tracks
    FILE *c_log_file



from mpf.config_players.plugin_player import PluginPlayer


class MpfTrackPlayer(PluginPlayer):
    """Base class for part of the track player which runs as part of MPF.

    Note: This class is loaded by MPF and everything in it is in the context of
    MPF, not the mpf-mc. MPF finds this instance because the mpf-mc setup.py
    has the following entry_point configured:

        track_player=mpfmc.config_players.track_player:register_with_mpf

    """
    config_file_section = 'track_player'
    show_section = 'tracks'


player_cls = MpfTrackPlayer


def register_with_mpf(machine):
    """Registers the sound player plug-in with MPF"""
    return 'track', MpfTrackPlayer(machine)



# This test is a bit unique because it's in the MPF-MC package but it runs on
# the MPF side of things, and therefore it uses MPF's test case instead of
# MPF-MC's.

import logging
import os
import time
from unittest.mock import *

from mpf.tests.loop import TimeTravelLoop, TestClock

import mpfmc
from mpf.tests.MpfTestCase import MpfTestCase, TestMachineController
import mpf.core
from mpf.core.config_validator import ConfigValidator


class TestMpfPluginConfigPlayerValidation(MpfTestCase):

    def get_absolute_machine_path(self):
        return os.path.abspath(os.path.join(
            mpfmc.__path__[0], os.pardir, 'mpfmc', self.get_machine_path()))

    def get_enable_plugins(self):
        return True

    def get_config_file(self):
        return 'mpf_plugin_validation.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/mpf_plugin_config_player_validation/'

    def test_slides_validation_shows(self):

        # for step in self.machine.shows['show1'].show_steps:
        #     print(step)
        #     print()

        self.assertIn('show1', self.machine.shows)

        step_0 = self.machine.shows['show1'].show_steps[0]['slides']['slide1']

        self.assertIn('widgets', step_0)
        self.assertEqual(1, len(step_0['widgets']))
        self.assertEqual('text', step_0['widgets'][0]['type'])
        self.assertEqual('TEST 1', step_0['widgets'][0]['text'])
        self.assertEqual([1.0, 0, 0, 1.0], step_0['widgets'][0]['color'])
        self.assertEqual(100, step_0['widgets'][0]['font_size'])

        self.machine.events.post('event1')
        self.advance_time_and_run(4)

        # for command in self.sent_bcp_commands:
        #     print()
        #     print(command)



from kivy.uix.effectwidget import EffectBase


class FlipVerticalEffect(EffectBase):
    """GLSL effect to vertically flip a texture"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def do_glsl(self):
        self.glsl = flip_vertical_glsl


flip_vertical_glsl = '''
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    return texture2D(texture, vec2(tex_coords.x, 1.0 - tex_coords.y));
}}
'''

effect_cls = FlipVerticalEffect
name = 'flip_vertical'



"""Test high score mode."""
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase

from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase


class TestDMDs(MpfIntegrationTestCase, MpfSlideTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/dmd/'

    def test_empty_name(self):
        default = self.machine.rgb_dmds["default"]
        grb = self.machine.rgb_dmds["grb_dmd"]
        self.advance_time_and_run(.1)
        self.assertEqual(bytearray([00] * 128 * 32 * 3), default.hw_device.data)
        self.assertEqual(bytearray([00] * 128 * 32 * 3), grb.hw_device.data)
        self.post_event("show_dmd_slide_1")
        self.advance_time_and_run(.1)
        self.assertEqual(bytearray([0x3, 0xe, 0x22] * 128 * 32), default.hw_device.data)
        self.assertEqual(bytearray([0xe, 0x3, 0x22] * 128 * 32), grb.hw_device.data)
